#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import uuid

from openstackclient.tests.functional.volume.v1 import common


class VolumeSnapshotTests(common.BaseVolumeTests):
    """Functional tests for volume snapshot. """

    VOLLY = uuid.uuid4().hex

    @classmethod
    def setUpClass(cls):
        super(VolumeSnapshotTests, cls).setUpClass()
        # create a volume for all tests to create snapshot
        cmd_output = json.loads(cls.openstack(
            'volume create -f json ' +
            '--size 1 ' +
            cls.VOLLY
        ))
        cls.wait_for_status('volume', cls.VOLLY, 'available')
        cls.VOLUME_ID = cmd_output['id']

    @classmethod
    def tearDownClass(cls):
        try:
            cls.wait_for_status('volume', cls.VOLLY, 'available')
            raw_output = cls.openstack('volume delete --force ' + cls.VOLLY)
            cls.assertOutput('', raw_output)
        finally:
            super(VolumeSnapshotTests, cls).tearDownClass()

    def test_volume_snapshot_delete(self):
        """Test create, delete multiple"""
        name1 = uuid.uuid4().hex
        cmd_output = json.loads(self.openstack(
            'volume snapshot create -f json ' +
            name1 +
            ' --volume ' + self.VOLLY
        ))
        self.assertEqual(
            name1,
            cmd_output["display_name"],
        )

        name2 = uuid.uuid4().hex
        cmd_output = json.loads(self.openstack(
            'volume snapshot create -f json ' +
            name2 +
            ' --volume ' + self.VOLLY
        ))
        self.assertEqual(
            name2,
            cmd_output["display_name"],
        )

        self.wait_for_status('volume snapshot', name1, 'available')
        self.wait_for_status('volume snapshot', name2, 'available')

        del_output = self.openstack(
            'volume snapshot delete ' + name1 + ' ' + name2)
        self.assertOutput('', del_output)
        self.wait_for_delete('volume snapshot', name1)
        self.wait_for_delete('volume snapshot', name2)

    def test_volume_snapshot_list(self):
        """Test create, list filter"""
        name1 = uuid.uuid4().hex
        cmd_output = json.loads(self.openstack(
            'volume snapshot create -f json ' +
            name1 +
            ' --volume ' + self.VOLLY
        ))
        self.addCleanup(self.wait_for_delete, 'volume snapshot', name1)
        self.addCleanup(self.openstack, 'volume snapshot delete ' + name1)
        self.assertEqual(
            name1,
            cmd_output["display_name"],
        )
        self.assertEqual(
            self.VOLUME_ID,
            cmd_output["volume_id"],
        )
        self.assertEqual(
            1,
            cmd_output["size"],
        )
        self.wait_for_status('volume snapshot', name1, 'available')

        name2 = uuid.uuid4().hex
        cmd_output = json.loads(self.openstack(
            'volume snapshot create -f json ' +
            name2 +
            ' --volume ' + self.VOLLY
        ))
        self.addCleanup(self.wait_for_delete, 'volume snapshot', name2)
        self.addCleanup(self.openstack, 'volume snapshot delete ' + name2)
        self.assertEqual(
            name2,
            cmd_output["display_name"],
        )
        self.assertEqual(
            self.VOLUME_ID,
            cmd_output["volume_id"],
        )
        self.assertEqual(
            1,
            cmd_output["size"],
        )
        self.wait_for_status('volume snapshot', name2, 'available')

        # Test list --long, --status
        cmd_output = json.loads(self.openstack(
            'volume snapshot list -f json ' +
            '--long ' +
            '--status error'
        ))
        names = [x["Name"] for x in cmd_output]
        self.assertNotIn(name1, names)
        self.assertNotIn(name2, names)

        # Test list --volume
        cmd_output = json.loads(self.openstack(
            'volume snapshot list -f json ' +
            '--volume ' + self.VOLLY
        ))
        names = [x["Name"] for x in cmd_output]
        self.assertIn(name1, names)
        self.assertIn(name2, names)

        # Test list --name
        cmd_output = json.loads(self.openstack(
            'volume snapshot list -f json ' +
            '--name ' + name1
        ))
        names = [x["Name"] for x in cmd_output]
        self.assertIn(name1, names)
        self.assertNotIn(name2, names)

    def test_snapshot_set(self):
        """Test create, set, unset, show, delete volume snapshot"""
        name = uuid.uuid4().hex
        new_name = name + "_"
        cmd_output = json.loads(self.openstack(
            'volume snapshot create -f json ' +
            '--volume ' + self.VOLLY +
            ' --description aaaa ' +
            name
        ))
        self.addCleanup(self.wait_for_delete, 'volume snapshot', new_name)
        self.addCleanup(self.openstack, 'volume snapshot delete ' + new_name)
        self.assertEqual(
            name,
            cmd_output["display_name"],
        )
        self.assertEqual(
            1,
            cmd_output["size"],
        )
        self.assertEqual(
            'aaaa',
            cmd_output["display_description"],
        )
        self.wait_for_status('volume snapshot', name, 'available')

        # Test volume snapshot set
        raw_output = self.openstack(
            'volume snapshot set ' +
            '--name ' + new_name +
            ' --description bbbb ' +
            '--property Alpha=a ' +
            '--property Beta=b ' +
            name,
        )
        self.assertOutput('', raw_output)

        # Show snapshot set result
        cmd_output = json.loads(self.openstack(
            'volume snapshot show -f json ' +
            new_name
        ))
        self.assertEqual(
            new_name,
            cmd_output["display_name"],
        )
        self.assertEqual(
            1,
            cmd_output["size"],
        )
        self.assertEqual(
            'bbbb',
            cmd_output["display_description"],
        )
        self.assertEqual(
            {'Alpha': 'a', 'Beta': 'b'},
            cmd_output["properties"],
        )

        # Test volume unset
        raw_output = self.openstack(
            'volume snapshot unset ' +
            '--property Alpha ' +
            new_name,
        )
        self.assertOutput('', raw_output)

        cmd_output = json.loads(self.openstack(
            'volume snapshot show -f json ' +
            new_name
        ))
        self.assertEqual(
            {'Beta': 'b'},
            cmd_output["properties"],
        )

        # Test volume snapshot set --no-property
        raw_output = self.openstack(
            'volume snapshot set ' +
            '--no-property ' +
            new_name,
        )
        self.assertOutput('', raw_output)
        cmd_output = json.loads(self.openstack(
            'volume snapshot show -f json ' +
            new_name
        ))
        self.assertEqual({}, cmd_output["properties"])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import uuid

from openstackclient.tests.functional.network.v2 import common


class PortTests(common.NetworkTagTests):
    """Functional tests for port"""

    base_command = 'port'

    NAME = uuid.uuid4().hex
    NETWORK_NAME = uuid.uuid4().hex

    @classmethod
    def setUpClass(cls):
        common.NetworkTests.setUpClass()
        if cls.haz_network:
            cls.NAME = uuid.uuid4().hex
            cls.NETWORK_NAME = uuid.uuid4().hex

            # Create a network for the port tests
            cls.openstack(
                'network create %s' % cls.NETWORK_NAME
            )

    @classmethod
    def tearDownClass(cls):
        try:
            if cls.haz_network:
                raw_output = cls.openstack(
                    'network delete %s' % cls.NETWORK_NAME
                )
                cls.assertOutput('', raw_output)
        finally:
            super(PortTests, cls).tearDownClass()

    def setUp(self):
        super(PortTests, self).setUp()
        # Nothing in this class works with Nova Network
        if not self.haz_network:
            self.skipTest("No Network service present")

    def test_port_delete(self):
        """Test create, delete multiple"""
        json_output = json.loads(self.openstack(
            'port create -f json --network %s %s' %
            (self.NETWORK_NAME, self.NAME)
        ))
        id1 = json_output.get('id')
        self.assertIsNotNone(id1)
        self.assertIsNotNone(json_output.get('mac_address'))
        self.assertEqual(self.NAME, json_output.get('name'))

        json_output = json.loads(self.openstack(
            'port create -f json --network %s %sx' %
            (self.NETWORK_NAME, self.NAME)
        ))
        id2 = json_output.get('id')
        self.assertIsNotNone(id2)
        self.assertIsNotNone(json_output.get('mac_address'))
        self.assertEqual(self.NAME + 'x', json_output.get('name'))

        # Clean up after ourselves
        raw_output = self.openstack('port delete %s %s' % (id1, id2))
        self.assertOutput('', raw_output)

    def test_port_list(self):
        """Test create defaults, list, delete"""
        json_output = json.loads(self.openstack(
            'port create -f json --network %s %s' %
            (self.NETWORK_NAME, self.NAME)
        ))
        id1 = json_output.get('id')
        self.assertIsNotNone(id1)
        mac1 = json_output.get('mac_address')
        self.assertIsNotNone(mac1)
        self.addCleanup(self.openstack, 'port delete %s' % id1)
        self.assertEqual(self.NAME, json_output.get('name'))

        json_output = json.loads(self.openstack(
            'port create -f json --network %s %sx' %
            (self.NETWORK_NAME, self.NAME)
        ))
        id2 = json_output.get('id')
        self.assertIsNotNone(id2)
        mac2 = json_output.get('mac_address')
        self.assertIsNotNone(mac2)
        self.addCleanup(self.openstack, 'port delete %s' % id2)
        self.assertEqual(self.NAME + 'x', json_output.get('name'))

        # Test list
        json_output = json.loads(self.openstack(
            'port list -f json'
        ))
        item_map = {item.get('ID'): item.get('MAC Address') for item in
                    json_output}
        self.assertIn(id1, item_map.keys())
        self.assertIn(id2, item_map.keys())
        self.assertIn(mac1, item_map.values())
        self.assertIn(mac2, item_map.values())

        # Test list --long
        json_output = json.loads(self.openstack(
            'port list --long -f json'
        ))
        id_list = [item.get('ID') for item in json_output]
        self.assertIn(id1, id_list)
        self.assertIn(id2, id_list)

        # Test list --mac-address
        json_output = json.loads(self.openstack(
            'port list -f json --mac-address %s' % mac2
        ))
        item_map = {item.get('ID'): item.get('MAC Address') for item in
                    json_output}
        self.assertNotIn(id1, item_map.keys())
        self.assertIn(id2, item_map.keys())
        self.assertNotIn(mac1, item_map.values())
        self.assertIn(mac2, item_map.values())

        # Test list with unknown fields
        json_output = json.loads(self.openstack(
            'port list -f json -c ID -c Name -c device_id'
        ))
        id_list = [p['ID'] for p in json_output]
        self.assertIn(id1, id_list)
        self.assertIn(id2, id_list)
        # Check an unknown field exists
        self.assertIn('device_id', json_output[0])

    def test_port_set(self):
        """Test create, set, show, delete"""
        name = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            'port create -f json '
            '--network %s '
            '--description xyzpdq '
            '--disable %s' %
            (self.NETWORK_NAME, name)
        ))
        id1 = json_output.get('id')
        self.addCleanup(self.openstack, 'port delete %s' % id1)
        self.assertEqual(name, json_output.get('name'))
        self.assertEqual('xyzpdq', json_output.get('description'))
        self.assertEqual(False, json_output.get('admin_state_up'))

        raw_output = self.openstack(
            'port set --enable %s' %
            name
        )
        self.assertOutput('', raw_output)

        json_output = json.loads(self.openstack(
            'port show -f json %s' % name
        ))
        sg_id = json_output.get('security_group_ids')[0]

        self.assertEqual(name, json_output.get('name'))
        self.assertEqual('xyzpdq', json_output.get('description'))
        self.assertEqual(True, json_output.get('admin_state_up'))
        self.assertIsNotNone(json_output.get('mac_address'))

        raw_output = self.openstack(
            'port unset --security-group %s %s' % (sg_id, id1))
        self.assertOutput('', raw_output)

        json_output = json.loads(self.openstack(
            'port show -f json %s' % name
        ))
        self.assertEqual([], json_output.get('security_group_ids'))

    def test_port_admin_set(self):
        """Test create, set (as admin), show, delete"""
        json_output = json.loads(self.openstack(
            'port create -f json '
            '--network %s %s' % (self.NETWORK_NAME, self.NAME)
        ))
        id_ = json_output.get('id')
        self.addCleanup(self.openstack, 'port delete %s' % id_)

        raw_output = self.openstack(
            '--os-username admin '
            'port set --mac-address 11:22:33:44:55:66 %s' %
            self.NAME)
        self.assertOutput('', raw_output)
        json_output = json.loads(self.openstack(
            'port show -f json %s' % self.NAME
        ))
        self.assertEqual(json_output.get('mac_address'), '11:22:33:44:55:66')

    def test_port_set_sg(self):
        """Test create, set, show, delete"""
        sg_name1 = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            'security group create -f json %s' %
            sg_name1
        ))
        sg_id1 = json_output.get('id')
        self.addCleanup(self.openstack, 'security group delete %s' % sg_id1)

        sg_name2 = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            'security group create -f json %s' %
            sg_name2
        ))
        sg_id2 = json_output.get('id')
        self.addCleanup(self.openstack, 'security group delete %s' % sg_id2)

        name = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            'port create -f json '
            '--network %s '
            '--security-group %s %s' %
            (self.NETWORK_NAME, sg_name1, name)
        ))
        id1 = json_output.get('id')
        self.addCleanup(self.openstack, 'port delete %s' % id1)
        self.assertEqual(name, json_output.get('name'))
        self.assertEqual([sg_id1], json_output.get('security_group_ids'))

        raw_output = self.openstack(
            'port set '
            '--security-group %s %s' %
            (sg_name2, name)
        )
        self.assertOutput('', raw_output)

        json_output = json.loads(self.openstack(
            'port show -f json %s' % name
        ))
        self.assertEqual(name, json_output.get('name'))
        # NOTE(amotoki): The order of the field is not predictable,
        self.assertIsInstance(json_output.get('security_group_ids'), list)
        self.assertEqual(sorted([sg_id1, sg_id2]),
                         sorted(json_output.get('security_group_ids')))

        raw_output = self.openstack(
            'port unset --security-group %s %s' % (sg_id1, id1))
        self.assertOutput('', raw_output)

        json_output = json.loads(self.openstack(
            'port show -f json %s' % name
        ))
        self.assertEqual(
            [sg_id2],
            json_output.get('security_group_ids')
        )

    def _create_resource_for_tag_test(self, name, args):
        return json.loads(self.openstack(
            '{} create -f json --network {} {} {}'
            .format(self.base_command, self.NETWORK_NAME, args, name)
        ))



#   Copyright 2019 SUSE LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
from unittest import mock

from osc_lib import exceptions
from osc_lib import utils

from openstackclient.identity.v3 import access_rule
from openstackclient.tests.unit import fakes
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestAccessRule(identity_fakes.TestIdentityv3):

    def setUp(self):
        super(TestAccessRule, self).setUp()

        identity_manager = self.app.client_manager.identity
        self.access_rules_mock = identity_manager.access_rules
        self.access_rules_mock.reset_mock()
        self.roles_mock = identity_manager.roles
        self.roles_mock.reset_mock()


class TestAccessRuleDelete(TestAccessRule):

    def setUp(self):
        super(TestAccessRuleDelete, self).setUp()

        # This is the return value for utils.find_resource()
        self.access_rules_mock.get.return_value = fakes.FakeResource(
            None,
            copy.deepcopy(identity_fakes.ACCESS_RULE),
            loaded=True,
        )
        self.access_rules_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = access_rule.DeleteAccessRule(
            self.app, None)

    def test_access_rule_delete(self):
        arglist = [
            identity_fakes.access_rule_id,
        ]
        verifylist = [
            ('access_rule', [identity_fakes.access_rule_id])
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.access_rules_mock.delete.assert_called_with(
            identity_fakes.access_rule_id,
        )
        self.assertIsNone(result)

    @mock.patch.object(utils, 'find_resource')
    def test_delete_multi_access_rules_with_exception(self, find_mock):
        find_mock.side_effect = [self.access_rules_mock.get.return_value,
                                 exceptions.CommandError]
        arglist = [
            identity_fakes.access_rule_id,
            'nonexistent_access_rule',
        ]
        verifylist = [
            ('access_rule', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 access rules failed to'
                             ' delete.', str(e))

        find_mock.assert_any_call(self.access_rules_mock,
                                  identity_fakes.access_rule_id)
        find_mock.assert_any_call(self.access_rules_mock,
                                  'nonexistent_access_rule')

        self.assertEqual(2, find_mock.call_count)
        self.access_rules_mock.delete.assert_called_once_with(
            identity_fakes.access_rule_id)


class TestAccessRuleList(TestAccessRule):

    def setUp(self):
        super(TestAccessRuleList, self).setUp()

        self.access_rules_mock.list.return_value = [
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ACCESS_RULE),
                loaded=True,
            ),
        ]

        # Get the command object to test
        self.cmd = access_rule.ListAccessRule(self.app, None)

    def test_access_rule_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.access_rules_mock.list.assert_called_with(user=None)

        collist = ('ID', 'Service', 'Method', 'Path')
        self.assertEqual(collist, columns)
        datalist = ((
            identity_fakes.access_rule_id,
            identity_fakes.access_rule_service,
            identity_fakes.access_rule_method,
            identity_fakes.access_rule_path,
        ), )
        self.assertEqual(datalist, tuple(data))


class TestAccessRuleShow(TestAccessRule):

    def setUp(self):
        super(TestAccessRuleShow, self).setUp()

        self.access_rules_mock.get.return_value = fakes.FakeResource(
            None,
            copy.deepcopy(identity_fakes.ACCESS_RULE),
            loaded=True,
        )

        # Get the command object to test
        self.cmd = access_rule.ShowAccessRule(self.app, None)

    def test_access_rule_show(self):
        arglist = [
            identity_fakes.access_rule_id,
        ]
        verifylist = [
            ('access_rule', identity_fakes.access_rule_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.access_rules_mock.get.assert_called_with(
            identity_fakes.access_rule_id)

        collist = ('id', 'method', 'path', 'service')
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.access_rule_id,
            identity_fakes.access_rule_method,
            identity_fakes.access_rule_path,
            identity_fakes.access_rule_service,
        )
        self.assertEqual(datalist, data)



#    Copyright 2018 SUSE Linux GmbH
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v3 import common


class ApplicationCredentialTests(common.IdentityTests):

    APPLICATION_CREDENTIAL_FIELDS = ['id', 'name', 'project_id',
                                     'description', 'roles', 'expires_at',
                                     'unrestricted']
    APPLICATION_CREDENTIAL_LIST_HEADERS = ['ID', 'Name', 'Project ID',
                                           'Description', 'Expires At']

    def test_application_credential_create(self):
        name = data_utils.rand_name('name')
        raw_output = self.openstack('application credential create %(name)s'
                                    % {'name': name})
        self.addCleanup(
            self.openstack,
            'application credential delete %(name)s' % {'name': name})
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.APPLICATION_CREDENTIAL_FIELDS)

    def _create_role_assignments(self):
        try:
            user = self.openstack('configuration show -f value'
                                  ' -c auth.username')
        except Exception:
            user = self.openstack('configuration show -f value'
                                  ' -c auth.user_id')
        try:
            user_domain = self.openstack('configuration show -f value'
                                         ' -c auth.user_domain_name')
        except Exception:
            user_domain = self.openstack('configuration show -f value'
                                         ' -c auth.user_domain_id')
        try:
            project = self.openstack('configuration show -f value'
                                     ' -c auth.project_name')
        except Exception:
            project = self.openstack('configuration show -f value'
                                     ' -c auth.project_id')
        try:
            project_domain = self.openstack('configuration show -f value'
                                            ' -c auth.project_domain_name')
        except Exception:
            project_domain = self.openstack('configuration show -f value'
                                            ' -c auth.project_domain_id')
        role1 = self._create_dummy_role()
        role2 = self._create_dummy_role()
        for role in role1, role2:
            self.openstack('role add'
                           ' --user %(user)s'
                           ' --user-domain %(user_domain)s'
                           ' --project %(project)s'
                           ' --project-domain %(project_domain)s'
                           ' %(role)s'
                           % {'user': user,
                              'user_domain': user_domain,
                              'project': project,
                              'project_domain': project_domain,
                              'role': role})
            self.addCleanup(self.openstack,
                            'role remove'
                            ' --user %(user)s'
                            ' --user-domain %(user_domain)s'
                            ' --project %(project)s'
                            ' --project-domain %(project_domain)s'
                            ' %(role)s'
                            % {'user': user,
                               'user_domain': user_domain,
                               'project': project,
                               'project_domain': project_domain,
                               'role': role})
        return role1, role2

    def test_application_credential_create_with_options(self):
        name = data_utils.rand_name('name')
        secret = data_utils.rand_name('secret')
        description = data_utils.rand_name('description')
        tomorrow = (datetime.datetime.utcnow() +
                    datetime.timedelta(days=1)).strftime('%Y-%m-%dT%H:%M:%S%z')
        role1, role2 = self._create_role_assignments()
        raw_output = self.openstack('application credential create %(name)s'
                                    ' --secret %(secret)s'
                                    ' --description %(description)s'
                                    ' --expiration %(tomorrow)s'
                                    ' --role %(role1)s'
                                    ' --role %(role2)s'
                                    ' --unrestricted'
                                    % {'name': name,
                                       'secret': secret,
                                       'description': description,
                                       'tomorrow': tomorrow,
                                       'role1': role1,
                                       'role2': role2})
        self.addCleanup(
            self.openstack,
            'application credential delete %(name)s' % {'name': name})
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.APPLICATION_CREDENTIAL_FIELDS)

    def test_application_credential_delete(self):
        name = data_utils.rand_name('name')
        self.openstack('application credential create %(name)s'
                       % {'name': name})
        raw_output = self.openstack('application credential delete '
                                    '%(name)s' % {'name': name})
        self.assertEqual(0, len(raw_output))

    def test_application_credential_list(self):
        raw_output = self.openstack('application credential list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(
            items, self.APPLICATION_CREDENTIAL_LIST_HEADERS)

    def test_application_credential_show(self):
        name = data_utils.rand_name('name')
        raw_output = self.openstack('application credential create %(name)s'
                                    % {'name': name})
        self.addCleanup(
            self.openstack,
            'application credential delete %(name)s' % {'name': name})
        raw_output = self.openstack('application credential show '
                                    '%(name)s' % {'name': name})
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.APPLICATION_CREDENTIAL_FIELDS)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume V2 Volume action implementations"""

import argparse
import copy
import functools
import logging

from cliff import columns as cliff_columns
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)


class AttachmentsColumn(cliff_columns.FormattableColumn):
    """Formattable column for attachments column.

    Unlike the parent FormattableColumn class, the initializer of the
    class takes server_cache as the second argument.
    osc_lib.utils.get_item_properties instantiate cliff FormattableColumn
    object with a single parameter "column value", so you need to pass
    a partially initialized class like
    ``functools.partial(AttachmentsColumn, server_cache)``.
    """

    def __init__(self, value, server_cache=None):
        super(AttachmentsColumn, self).__init__(value)
        self._server_cache = server_cache or {}

    def human_readable(self):
        """Return a formatted string of a volume's attached instances

        :rtype: a string of formatted instances
        """

        msg = ''
        for attachment in self._value:
            server = attachment['server_id']
            if server in self._server_cache.keys():
                server = self._server_cache[server].name
            device = attachment['device']
            msg += 'Attached to %s on %s ' % (server, device)
        return msg


def _check_size_arg(args):
    """Check whether --size option is required or not.

    Require size parameter only in case when snapshot or source
    volume is not specified.
    """

    if ((args.snapshot or args.source)
            is None and args.size is None):
        msg = _("--size is a required option if snapshot "
                "or source volume is not specified.")
        raise exceptions.CommandError(msg)


class CreateVolume(command.ShowOne):
    _description = _("Create new volume")

    def get_parser(self, prog_name):
        parser = super(CreateVolume, self).get_parser(prog_name)
        parser.add_argument(
            "name",
            metavar="<name>",
            help=_("Volume name"),
        )
        parser.add_argument(
            "--size",
            metavar="<size>",
            type=int,
            help=_("Volume size in GB (Required unless --snapshot or "
                   "--source is specified)"),
        )
        parser.add_argument(
            "--type",
            metavar="<volume-type>",
            help=_("Set the type of volume"),
        )
        source_group = parser.add_mutually_exclusive_group()
        source_group.add_argument(
            "--image",
            metavar="<image>",
            help=_("Use <image> as source of volume (name or ID)"),
        )
        source_group.add_argument(
            "--snapshot",
            metavar="<snapshot>",
            help=_("Use <snapshot> as source of volume (name or ID)"),
        )
        source_group.add_argument(
            "--source",
            metavar="<volume>",
            help=_("Volume to clone (name or ID)"),
        )
        source_group.add_argument(
            "--source-replicated",
            metavar="<replicated-volume>",
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Volume description"),
        )
        parser.add_argument(
            "--availability-zone",
            metavar="<availability-zone>",
            help=_("Create volume in <availability-zone>"),
        )
        parser.add_argument(
            "--consistency-group",
            metavar="consistency-group>",
            help=_("Consistency group where the new volume belongs to"),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            help=_("Set a property to this volume "
                   "(repeat option to set multiple properties)"),
        )
        parser.add_argument(
            "--hint",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            help=_("Arbitrary scheduler hint key-value pairs to help boot "
                   "an instance (repeat option to set multiple hints)"),
        )
        bootable_group = parser.add_mutually_exclusive_group()
        bootable_group.add_argument(
            "--bootable",
            action="store_true",
            help=_("Mark volume as bootable")
        )
        bootable_group.add_argument(
            "--non-bootable",
            action="store_true",
            help=_("Mark volume as non-bootable (default)")
        )
        readonly_group = parser.add_mutually_exclusive_group()
        readonly_group.add_argument(
            "--read-only",
            action="store_true",
            help=_("Set volume to read-only access mode")
        )
        readonly_group.add_argument(
            "--read-write",
            action="store_true",
            help=_("Set volume to read-write access mode (default)")
        )
        return parser

    def take_action(self, parsed_args):
        _check_size_arg(parsed_args)
        volume_client = self.app.client_manager.volume
        image_client = self.app.client_manager.image

        source_volume = None
        if parsed_args.source:
            source_volume = utils.find_resource(
                volume_client.volumes,
                parsed_args.source).id

        consistency_group = None
        if parsed_args.consistency_group:
            consistency_group = utils.find_resource(
                volume_client.consistencygroups,
                parsed_args.consistency_group).id

        image = None
        if parsed_args.image:
            image = image_client.find_image(parsed_args.image,
                                            ignore_missing=False).id

        size = parsed_args.size

        snapshot = None
        if parsed_args.snapshot:
            snapshot_obj = utils.find_resource(
                volume_client.volume_snapshots,
                parsed_args.snapshot)
            snapshot = snapshot_obj.id
            # Cinder requires a value for size when creating a volume
            # even if creating from a snapshot. Cinder will create the
            # volume with at least the same size as the snapshot anyway,
            # so since we have the object here, just override the size
            # value if it's either not given or is smaller than the
            # snapshot size.
            size = max(size or 0, snapshot_obj.size)

        volume = volume_client.volumes.create(
            size=size,
            snapshot_id=snapshot,
            name=parsed_args.name,
            description=parsed_args.description,
            volume_type=parsed_args.type,
            availability_zone=parsed_args.availability_zone,
            metadata=parsed_args.property,
            imageRef=image,
            source_volid=source_volume,
            consistencygroup_id=consistency_group,
            scheduler_hints=parsed_args.hint,
        )

        if parsed_args.bootable or parsed_args.non_bootable:
            try:
                volume_client.volumes.set_bootable(
                    volume.id, parsed_args.bootable)
            except Exception as e:
                LOG.error(_("Failed to set volume bootable property: %s"), e)
        if parsed_args.read_only or parsed_args.read_write:
            try:
                volume_client.volumes.update_readonly_flag(
                    volume.id,
                    parsed_args.read_only)
            except Exception as e:
                LOG.error(_("Failed to set volume read-only access "
                            "mode flag: %s"), e)

        # Remove key links from being displayed
        volume._info.update(
            {
                'properties':
                format_columns.DictColumn(volume._info.pop('metadata')),
                'type': volume._info.pop('volume_type')
            }
        )
        volume._info.pop("links", None)
        return zip(*sorted(volume._info.items()))


class DeleteVolume(command.Command):
    _description = _("Delete volume(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteVolume, self).get_parser(prog_name)
        parser.add_argument(
            "volumes",
            metavar="<volume>",
            nargs="+",
            help=_("Volume(s) to delete (name or ID)")
        )
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--force",
            action="store_true",
            help=_("Attempt forced removal of volume(s), regardless of state "
                   "(defaults to False)")
        )
        group.add_argument(
            "--purge",
            action="store_true",
            help=_("Remove any snapshots along with volume(s) "
                   "(defaults to False)")
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        result = 0

        for i in parsed_args.volumes:
            try:
                volume_obj = utils.find_resource(
                    volume_client.volumes, i)
                if parsed_args.force:
                    volume_client.volumes.force_delete(volume_obj.id)
                else:
                    volume_client.volumes.delete(volume_obj.id,
                                                 cascade=parsed_args.purge)
            except Exception as e:
                result += 1
                LOG.error(_("Failed to delete volume with "
                            "name or ID '%(volume)s': %(e)s"),
                          {'volume': i, 'e': e})

        if result > 0:
            total = len(parsed_args.volumes)
            msg = (_("%(result)s of %(total)s volumes failed "
                   "to delete.") % {'result': result, 'total': total})
            raise exceptions.CommandError(msg)


class ListVolume(command.Lister):
    _description = _("List volumes")

    def get_parser(self, prog_name):
        parser = super(ListVolume, self).get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Filter results by project (name or ID) (admin only)')
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Filter results by user (name or ID) (admin only)')
        )
        identity_common.add_user_domain_option_to_parser(parser)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Filter results by volume name'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            help=_('Filter results by status'),
        )
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            '--marker',
            metavar='<volume>',
            help=_('The last volume ID of the previous page'),
        )
        parser.add_argument(
            '--limit',
            type=int,
            action=parseractions.NonNegativeAction,
            metavar='<num-volumes>',
            help=_('Maximum number of volumes to display'),
        )
        return parser

    def take_action(self, parsed_args):

        volume_client = self.app.client_manager.volume
        identity_client = self.app.client_manager.identity

        if parsed_args.long:
            columns = [
                'ID',
                'Name',
                'Status',
                'Size',
                'Volume Type',
                'Bootable',
                'Attachments',
                'Metadata',
            ]
            column_headers = copy.deepcopy(columns)
            column_headers[4] = 'Type'
            column_headers[6] = 'Attached to'
            column_headers[7] = 'Properties'
        else:
            columns = [
                'ID',
                'Name',
                'Status',
                'Size',
                'Attachments',
            ]
            column_headers = copy.deepcopy(columns)
            column_headers[4] = 'Attached to'

        # Cache the server list
        server_cache = {}
        try:
            compute_client = self.app.client_manager.compute
            for s in compute_client.servers.list():
                server_cache[s.id] = s
        except Exception:
            # Just forget it if there's any trouble
            pass
        AttachmentsColumnWithCache = functools.partial(
            AttachmentsColumn, server_cache=server_cache)

        project_id = None
        if parsed_args.project:
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain).id

        user_id = None
        if parsed_args.user:
            user_id = identity_common.find_user(identity_client,
                                                parsed_args.user,
                                                parsed_args.user_domain).id

        # set value of 'all_tenants' when using project option
        all_projects = bool(parsed_args.project) or parsed_args.all_projects

        search_opts = {
            'all_tenants': all_projects,
            'project_id': project_id,
            'user_id': user_id,
            'name': parsed_args.name,
            'status': parsed_args.status,
        }

        data = volume_client.volumes.list(
            search_opts=search_opts,
            marker=parsed_args.marker,
            limit=parsed_args.limit,
        )
        column_headers = utils.backward_compat_col_lister(
            column_headers, parsed_args.columns, {'Display Name': 'Name'})

        return (column_headers,
                (utils.get_item_properties(
                    s, columns,
                    formatters={'Metadata': format_columns.DictColumn,
                                'Attachments': AttachmentsColumnWithCache},
                ) for s in data))


class MigrateVolume(command.Command):
    _description = _("Migrate volume to a new host")

    def get_parser(self, prog_name):
        parser = super(MigrateVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar="<volume>",
            help=_("Volume to migrate (name or ID)")
        )
        parser.add_argument(
            '--host',
            metavar="<host>",
            required=True,
            help=_("Destination host (takes the form: host@backend-name#pool)")
        )
        parser.add_argument(
            '--force-host-copy',
            action="store_true",
            help=_("Enable generic host-based force-migration, "
                   "which bypasses driver optimizations")
        )
        parser.add_argument(
            '--lock-volume',
            action="store_true",
            help=_("If specified, the volume state will be locked "
                   "and will not allow a migration to be aborted "
                   "(possibly by another operation)")
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume = utils.find_resource(volume_client.volumes, parsed_args.volume)
        volume_client.volumes.migrate_volume(volume.id, parsed_args.host,
                                             parsed_args.force_host_copy,
                                             parsed_args.lock_volume,)


class SetVolume(command.Command):
    _description = _("Set volume properties")

    def get_parser(self, prog_name):
        parser = super(SetVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to modify (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New volume name'),
        )
        parser.add_argument(
            '--size',
            metavar='<size>',
            type=int,
            help=_('Extend volume size in GB'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New volume description'),
        )
        parser.add_argument(
            "--no-property",
            dest="no_property",
            action="store_true",
            help=_("Remove all properties from <volume> "
                   "(specify both --no-property and --property to "
                   "remove the current properties before setting "
                   "new properties.)"),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            help=_('Set a property on this volume '
                   '(repeat option to set multiple properties)'),
        )
        parser.add_argument(
            '--image-property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            help=_('Set an image property on this volume '
                   '(repeat option to set multiple image properties)'),
        )
        parser.add_argument(
            "--state",
            metavar="<state>",
            choices=['available', 'error', 'creating', 'deleting',
                     'in-use', 'attaching', 'detaching', 'error_deleting',
                     'maintenance'],
            help=_('New volume state ("available", "error", "creating", '
                   '"deleting", "in-use", "attaching", "detaching", '
                   '"error_deleting" or "maintenance") (admin only) '
                   '(This option simply changes the state of the volume '
                   'in the database with no regard to actual status, '
                   'exercise caution when using)'),
        )
        attached_group = parser.add_mutually_exclusive_group()
        attached_group.add_argument(
            "--attached",
            action="store_true",
            help=_('Set volume attachment status to "attached" '
                   '(admin only) '
                   '(This option simply changes the state of the volume '
                   'in the database with no regard to actual status, '
                   'exercise caution when using)'),
        )
        attached_group.add_argument(
            "--detached",
            action="store_true",
            help=_('Set volume attachment status to "detached" '
                   '(admin only) '
                   '(This option simply changes the state of the volume '
                   'in the database with no regard to actual status, '
                   'exercise caution when using)'),
        )
        parser.add_argument(
            '--type',
            metavar='<volume-type>',
            help=_('New volume type (name or ID)'),
        )
        parser.add_argument(
            '--retype-policy',
            metavar='<retype-policy>',
            choices=['never', 'on-demand'],
            help=_('Migration policy while re-typing volume '
                   '("never" or "on-demand", default is "never" ) '
                   '(available only when --type option is specified)'),
        )
        bootable_group = parser.add_mutually_exclusive_group()
        bootable_group.add_argument(
            "--bootable",
            action="store_true",
            help=_("Mark volume as bootable")
        )
        bootable_group.add_argument(
            "--non-bootable",
            action="store_true",
            help=_("Mark volume as non-bootable")
        )
        readonly_group = parser.add_mutually_exclusive_group()
        readonly_group.add_argument(
            "--read-only",
            action="store_true",
            help=_("Set volume to read-only access mode")
        )
        readonly_group.add_argument(
            "--read-write",
            action="store_true",
            help=_("Set volume to read-write access mode")
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume = utils.find_resource(volume_client.volumes, parsed_args.volume)

        result = 0
        if parsed_args.size:
            try:
                if parsed_args.size <= volume.size:
                    msg = (_("New size must be greater than %s GB")
                           % volume.size)
                    raise exceptions.CommandError(msg)
                if volume.status != 'available' and \
                        not volume_client.api_version.matches('3.42'):

                    msg = (_("Volume is in %s state, it must be available "
                           "before size can be extended") % volume.status)
                    raise exceptions.CommandError(msg)
                volume_client.volumes.extend(volume.id, parsed_args.size)
            except Exception as e:
                LOG.error(_("Failed to set volume size: %s"), e)
                result += 1

        if parsed_args.no_property:
            try:
                volume_client.volumes.delete_metadata(
                    volume.id, volume.metadata.keys())
            except Exception as e:
                LOG.error(_("Failed to clean volume properties: %s"), e)
                result += 1

        if parsed_args.property:
            try:
                volume_client.volumes.set_metadata(
                    volume.id, parsed_args.property)
            except Exception as e:
                LOG.error(_("Failed to set volume property: %s"), e)
                result += 1
        if parsed_args.image_property:
            try:
                volume_client.volumes.set_image_metadata(
                    volume.id, parsed_args.image_property)
            except Exception as e:
                LOG.error(_("Failed to set image property: %s"), e)
                result += 1
        if parsed_args.state:
            try:
                volume_client.volumes.reset_state(
                    volume.id, parsed_args.state)
            except Exception as e:
                LOG.error(_("Failed to set volume state: %s"), e)
                result += 1
        if parsed_args.attached:
            try:
                volume_client.volumes.reset_state(
                    volume.id, state=None,
                    attach_status="attached")
            except Exception as e:
                LOG.error(_("Failed to set volume attach-status: %s"), e)
                result += 1
        if parsed_args.detached:
            try:
                volume_client.volumes.reset_state(
                    volume.id, state=None,
                    attach_status="detached")
            except Exception as e:
                LOG.error(_("Failed to set volume attach-status: %s"), e)
                result += 1
        if parsed_args.bootable or parsed_args.non_bootable:
            try:
                volume_client.volumes.set_bootable(
                    volume.id, parsed_args.bootable)
            except Exception as e:
                LOG.error(_("Failed to set volume bootable property: %s"), e)
                result += 1
        if parsed_args.read_only or parsed_args.read_write:
            try:
                volume_client.volumes.update_readonly_flag(
                    volume.id,
                    parsed_args.read_only)
            except Exception as e:
                LOG.error(_("Failed to set volume read-only access "
                            "mode flag: %s"), e)
                result += 1
        if parsed_args.type:
            # get the migration policy
            migration_policy = 'never'
            if parsed_args.retype_policy:
                migration_policy = parsed_args.retype_policy
            try:
                # find the volume type
                volume_type = utils.find_resource(
                    volume_client.volume_types,
                    parsed_args.type)
                # reset to the new volume type
                volume_client.volumes.retype(
                    volume.id,
                    volume_type.id,
                    migration_policy)
            except Exception as e:
                LOG.error(_("Failed to set volume type: %s"), e)
                result += 1
        elif parsed_args.retype_policy:
            # If the "--retype-policy" is specified without "--type"
            LOG.warning(_("'--retype-policy' option will not work "
                          "without '--type' option"))

        kwargs = {}
        if parsed_args.name:
            kwargs['display_name'] = parsed_args.name
        if parsed_args.description:
            kwargs['display_description'] = parsed_args.description
        if kwargs:
            try:
                volume_client.volumes.update(volume.id, **kwargs)
            except Exception as e:
                LOG.error(_("Failed to update volume display name "
                          "or display description: %s"), e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(_("One or more of the "
                                          "set operations failed"))


class ShowVolume(command.ShowOne):
    _description = _("Display volume details")

    def get_parser(self, prog_name):
        parser = super(ShowVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar="<volume>",
            help=_("Volume to display (name or ID)")
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume = utils.find_resource(volume_client.volumes, parsed_args.volume)

        # Special mapping for columns to make the output easier to read:
        # 'metadata' --> 'properties'
        # 'volume_type' --> 'type'
        volume._info.update(
            {
                'properties':
                format_columns.DictColumn(volume._info.pop('metadata')),
                'type': volume._info.pop('volume_type'),
            },
        )

        # Remove key links from being displayed
        volume._info.pop("links", None)
        return zip(*sorted(volume._info.items()))


class UnsetVolume(command.Command):
    _description = _("Unset volume properties")

    def get_parser(self, prog_name):
        parser = super(UnsetVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            help=_('Remove a property from volume '
                   '(repeat option to remove multiple properties)'),
        )
        parser.add_argument(
            '--image-property',
            metavar='<key>',
            action='append',
            help=_('Remove an image property from volume '
                   '(repeat option to remove multiple image properties)'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume = utils.find_resource(
            volume_client.volumes, parsed_args.volume)

        result = 0
        if parsed_args.property:
            try:
                volume_client.volumes.delete_metadata(
                    volume.id, parsed_args.property)
            except Exception as e:
                LOG.error(_("Failed to unset volume property: %s"), e)
                result += 1

        if parsed_args.image_property:
            try:
                volume_client.volumes.delete_image_metadata(
                    volume.id, parsed_args.image_property)
            except Exception as e:
                LOG.error(_("Failed to unset image property: %s"), e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(_("One or more of the "
                                          "unset operations failed"))



# Copyright 2020, Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Compute v2 Server action implementations"""

from novaclient import api_versions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _


class ListServerVolume(command.Lister):
    """List all the volumes attached to a server."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            help=_('Server to list volume attachments for (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):

        compute_client = self.app.client_manager.compute

        server = utils.find_resource(
            compute_client.servers,
            parsed_args.server,
        )

        volumes = compute_client.volumes.get_server_volumes(server.id)

        columns = ()
        column_headers = ()

        if compute_client.api_version < api_versions.APIVersion('2.89'):
            columns += ('id',)
            column_headers += ('ID',)

        columns += (
            'device',
            'serverId',
            'volumeId',
        )
        column_headers += (
            'Device',
            'Server ID',
            'Volume ID',
        )

        if compute_client.api_version >= api_versions.APIVersion('2.70'):
            columns += ('tag',)
            column_headers += ('Tag',)

        if compute_client.api_version >= api_versions.APIVersion('2.79'):
            columns += ('delete_on_termination',)
            column_headers += ('Delete On Termination?',)

        if compute_client.api_version >= api_versions.APIVersion('2.89'):
            columns += ('attachment_id', 'bdm_uuid')
            column_headers += ('Attachment ID', 'BlockDeviceMapping UUID')

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s, columns, mixed_case_fields=('serverId', 'volumeId')
                ) for s in volumes
            ),
        )


class UpdateServerVolume(command.Command):
    """Update a volume attachment on the server."""

    def get_parser(self, prog_name):
        parser = super(UpdateServerVolume, self).get_parser(prog_name)
        parser.add_argument(
            'server',
            help=_('Server to update volume for (name or ID)'),
        )
        parser.add_argument(
            'volume',
            help=_('Volume (ID)'),
        )
        termination_group = parser.add_mutually_exclusive_group()
        termination_group.add_argument(
            '--delete-on-termination',
            action='store_true',
            dest='delete_on_termination',
            default=None,
            help=_(
                'Delete the volume when the server is destroyed '
                '(supported by --os-compute-api-version 2.85 or above)'
            ),
        )
        termination_group.add_argument(
            '--preserve-on-termination',
            action='store_false',
            dest='delete_on_termination',
            help=_(
                'Preserve the volume when the server is destroyed '
                '(supported by --os-compute-api-version 2.85 or above)'
            ),
        )
        return parser

    def take_action(self, parsed_args):

        compute_client = self.app.client_manager.compute

        if parsed_args.delete_on_termination is not None:
            if compute_client.api_version < api_versions.APIVersion('2.85'):
                msg = _(
                    '--os-compute-api-version 2.85 or greater is required to '
                    'support the --(no-)delete-on-termination option'
                )
                raise exceptions.CommandError(msg)

            server = utils.find_resource(
                compute_client.servers,
                parsed_args.server,
            )

            # NOTE(stephenfin): This may look silly, and that's because it is.
            # This API was originally used only for the swapping volumes, which
            # is an internal operation that should only be done by
            # orchestration software rather than a human. We're not going to
            # expose that, but we are going to expose the ability to change the
            # delete on termination behavior.
            compute_client.volumes.update_server_volume(
                server.id,
                parsed_args.volume,
                parsed_args.volume,
                delete_on_termination=parsed_args.delete_on_termination,
            )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Image v1 API Library"""

from openstackclient.api import api


class APIv1(api.BaseAPI):
    """Image v1 API"""

    _endpoint_suffix = '/v1'

    def __init__(self, endpoint=None, **kwargs):
        super(APIv1, self).__init__(endpoint=endpoint, **kwargs)

        self.endpoint = self.endpoint.rstrip('/')
        self._munge_url()

    def _munge_url(self):
        # Hack this until discovery is up
        if not self.endpoint.endswith(self._endpoint_suffix):
            self.endpoint = self.endpoint + self._endpoint_suffix

    def image_list(
        self,
        detailed=False,
        public=False,
        private=False,
        **filter
    ):
        """Get available images

        :param detailed:
            Retrieve detailed response from server if True
        :param public:
            Return public images if True
        :param private:
            Return private images if True

        If public and private are both True or both False then all images are
        returned.  Both arguments False is equivalent to no filter and all
        images are returned.  Both arguments True is a filter that includes
        both public and private images which is the same set as all images.

        http://docs.openstack.org/api/openstack-image-service/1.1/content/requesting-a-list-of-public-vm-images.html
        http://docs.openstack.org/api/openstack-image-service/1.1/content/requesting-detailed-metadata-on-public-vm-images.html
        http://docs.openstack.org/api/openstack-image-service/1.1/content/filtering-images-returned-via-get-images-and-get-imagesdetail.html
        """

        url = "/images"
        if detailed or public or private:
            # Because we can't all use /details
            url += "/detail"

        image_list = self.list(url, **filter)['images']

        if public != private:
            # One is True and one is False, so public represents the filter
            # state in either case
            image_list = [i for i in image_list if i['is_public'] == public]

        return image_list



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from openstackclient.common import extension
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit import fakes
from openstackclient.tests.unit.identity.v2_0 import fakes as identity_fakes
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils
from openstackclient.tests.unit import utils as tests_utils
from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes


class TestExtension(utils.TestCommand):

    def setUp(self):
        super(TestExtension, self).setUp()

        identity_client = identity_fakes.FakeIdentityv2Client(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )
        self.app.client_manager.identity = identity_client
        self.identity_extensions_mock = identity_client.extensions
        self.identity_extensions_mock.reset_mock()

        sdk_connection = mock.Mock()
        self.app.client_manager.sdk_connection = sdk_connection
        self.compute_extensions_mock = sdk_connection.compute.extensions
        self.compute_extensions_mock.reset_mock()

        volume_client = volume_fakes.FakeVolumeClient(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )
        self.app.client_manager.volume = volume_client
        volume_client.list_extensions = mock.Mock()
        self.volume_extensions_mock = volume_client.list_extensions
        self.volume_extensions_mock.reset_mock()

        network_client = network_fakes.FakeNetworkV2Client(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )
        self.app.client_manager.network = network_client
        network_client.extensions = mock.Mock()
        self.network_extensions_mock = network_client.extensions
        self.network_extensions_mock.reset_mock()


class TestExtensionList(TestExtension):

    columns = ('Name', 'Alias', 'Description')
    long_columns = ('Name', 'Alias', 'Description', 'Namespace', 'Updated',
                    'Links')

    volume_extension = volume_fakes.FakeExtension.create_one_extension()
    identity_extension = identity_fakes.FakeExtension.create_one_extension()
    compute_extension = compute_fakes.FakeExtension.create_one_extension()
    network_extension = network_fakes.FakeExtension.create_one_extension()

    def setUp(self):
        super(TestExtensionList, self).setUp()

        self.identity_extensions_mock.list.return_value = [
            self.identity_extension]
        self.compute_extensions_mock.return_value = [self.compute_extension]
        self.volume_extensions_mock.show_all.return_value = [
            self.volume_extension]
        self.network_extensions_mock.return_value = [self.network_extension]

        # Get the command object to test
        self.cmd = extension.ListExtension(self.app, None)

    def _test_extension_list_helper(self, arglist, verifylist,
                                    expected_data, long=False):
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        if long:
            self.assertEqual(self.long_columns, columns)
        else:
            self.assertEqual(self.columns, columns)
        self.assertEqual(expected_data, tuple(data))

    def test_extension_list_no_options(self):
        arglist = []
        verifylist = []
        datalist = (
            (
                self.identity_extension.name,
                self.identity_extension.alias,
                self.identity_extension.description,
            ),
            (
                self.compute_extension.name,
                self.compute_extension.alias,
                self.compute_extension.description,
            ),
            (
                self.volume_extension.name,
                self.volume_extension.alias,
                self.volume_extension.description,
            ),
            (
                self.network_extension.name,
                self.network_extension.alias,
                self.network_extension.description,
            ),
        )
        self._test_extension_list_helper(arglist, verifylist, datalist)
        self.identity_extensions_mock.list.assert_called_with()
        self.compute_extensions_mock.assert_called_with()
        self.volume_extensions_mock.show_all.assert_called_with()
        self.network_extensions_mock.assert_called_with()

    def test_extension_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        datalist = (
            (
                self.identity_extension.name,
                self.identity_extension.alias,
                self.identity_extension.description,
                self.identity_extension.namespace,
                self.identity_extension.updated,
                self.identity_extension.links,
            ),
            (
                self.compute_extension.name,
                self.compute_extension.alias,
                self.compute_extension.description,
                self.compute_extension.namespace,
                self.compute_extension.updated,
                self.compute_extension.links,
            ),
            (
                self.volume_extension.name,
                self.volume_extension.alias,
                self.volume_extension.description,
                self.volume_extension.namespace,
                self.volume_extension.updated,
                self.volume_extension.links,
            ),
            (
                self.network_extension.name,
                self.network_extension.alias,
                self.network_extension.description,
                self.network_extension.namespace,
                self.network_extension.updated,
                self.network_extension.links,
            ),
        )
        self._test_extension_list_helper(arglist, verifylist, datalist, True)
        self.identity_extensions_mock.list.assert_called_with()
        self.compute_extensions_mock.assert_called_with()
        self.volume_extensions_mock.show_all.assert_called_with()
        self.network_extensions_mock.assert_called_with()

    def test_extension_list_identity(self):
        arglist = [
            '--identity',
        ]
        verifylist = [
            ('identity', True),
        ]
        datalist = ((
            self.identity_extension.name,
            self.identity_extension.alias,
            self.identity_extension.description,
        ), )
        self._test_extension_list_helper(arglist, verifylist, datalist)
        self.identity_extensions_mock.list.assert_called_with()

    def test_extension_list_network(self):
        arglist = [
            '--network',
        ]
        verifylist = [
            ('network', True),
        ]
        datalist = (
            (
                self.network_extension.name,
                self.network_extension.alias,
                self.network_extension.description,
            ),
        )
        self._test_extension_list_helper(arglist, verifylist, datalist)
        self.network_extensions_mock.assert_called_with()

    def test_extension_list_network_with_long(self):
        arglist = [
            '--network',
            '--long',
        ]
        verifylist = [
            ('network', True),
            ('long', True),
        ]
        datalist = ((
            self.network_extension.name,
            self.network_extension.alias,
            self.network_extension.description,
            self.network_extension.namespace,
            self.network_extension.updated,
            self.network_extension.links,
        ), )
        self._test_extension_list_helper(arglist, verifylist, datalist,
                                         long=True)
        self.network_extensions_mock.assert_called_with()

    def test_extension_list_compute(self):
        arglist = [
            '--compute',
        ]
        verifylist = [
            ('compute', True),
        ]
        datalist = ((
            self.compute_extension.name,
            self.compute_extension.alias,
            self.compute_extension.description,
        ), )
        self._test_extension_list_helper(arglist, verifylist, datalist)
        self.compute_extensions_mock.assert_called_with()

    def test_extension_list_compute_and_network(self):
        arglist = [
            '--compute',
            '--network',
        ]
        verifylist = [
            ('compute', True),
            ('network', True),
        ]
        datalist = (
            (
                self.compute_extension.name,
                self.compute_extension.alias,
                self.compute_extension.description,
            ),
            (
                self.network_extension.name,
                self.network_extension.alias,
                self.network_extension.description,
            ),
        )
        self._test_extension_list_helper(arglist, verifylist, datalist)
        self.compute_extensions_mock.assert_called_with()
        self.network_extensions_mock.assert_called_with()

    def test_extension_list_volume(self):
        arglist = [
            '--volume',
        ]
        verifylist = [
            ('volume', True),
        ]
        datalist = ((
            self.volume_extension.name,
            self.volume_extension.alias,
            self.volume_extension.description,
        ), )
        self._test_extension_list_helper(arglist, verifylist, datalist)
        self.volume_extensions_mock.show_all.assert_called_with()


class TestExtensionShow(TestExtension):
    extension_details = (
        network_fakes.FakeExtension.create_one_extension()
    )

    columns = (
        'alias',
        'description',
        'links',
        'name',
        'namespace',
        'updated',
    )

    data = (
        extension_details.alias,
        extension_details.description,
        extension_details.links,
        extension_details.name,
        extension_details.namespace,
        extension_details.updated
    )

    def setUp(self):
        super(TestExtensionShow, self).setUp()

        self.cmd = extension.ShowExtension(self.app, None)

        self.app.client_manager.network.find_extension = mock.Mock(
            return_value=self.extension_details)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(tests_utils.ParserException, self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_show_all_options(self):
        arglist = [
            self.extension_details.alias,
        ]
        verifylist = [
            ('extension', self.extension_details.alias),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.app.client_manager.network.find_extension.assert_called_with(
            self.extension_details.alias, ignore_missing=False)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Object v1 action implementations"""

import logging

from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class CreateObject(command.Lister):
    _description = _("Upload object to container")

    def get_parser(self, prog_name):
        parser = super(CreateObject, self).get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Container for new object'),
        )
        parser.add_argument(
            'objects',
            metavar='<filename>',
            nargs="+",
            help=_('Local filename(s) to upload'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Upload a file and rename it. '
                   'Can only be used when uploading a single object')
        )
        return parser

    def take_action(self, parsed_args):
        if parsed_args.name:
            if len(parsed_args.objects) > 1:
                msg = _('Attempting to upload multiple objects and '
                        'using --name is not permitted')
                raise exceptions.CommandError(msg)
        results = []
        for obj in parsed_args.objects:
            if len(obj) > 1024:
                LOG.warning(
                    _('Object name is %s characters long, default limit'
                      ' is 1024'), len(obj))
            data = self.app.client_manager.object_store.object_create(
                container=parsed_args.container,
                object=obj,
                name=parsed_args.name,
            )
            results.append(data)

        columns = ("object", "container", "etag")
        return (columns,
                (utils.get_dict_properties(
                    s, columns,
                    formatters={},
                ) for s in results))


class DeleteObject(command.Command):
    _description = _("Delete object from container")

    def get_parser(self, prog_name):
        parser = super(DeleteObject, self).get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Delete object(s) from <container>'),
        )
        parser.add_argument(
            'objects',
            metavar='<object>',
            nargs="+",
            help=_('Object(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args):

        for obj in parsed_args.objects:
            self.app.client_manager.object_store.object_delete(
                container=parsed_args.container,
                object=obj,
            )


class ListObject(command.Lister):
    _description = _("List objects")

    def get_parser(self, prog_name):
        parser = super(ListObject, self).get_parser(prog_name)
        parser.add_argument(
            "container",
            metavar="<container>",
            help=_("Container to list"),
        )
        parser.add_argument(
            "--prefix",
            metavar="<prefix>",
            help=_("Filter list using <prefix>"),
        )
        parser.add_argument(
            "--delimiter",
            metavar="<delimiter>",
            help=_("Roll up items with <delimiter>"),
        )
        parser.add_argument(
            "--marker",
            metavar="<marker>",
            help=_("Anchor for paging"),
        )
        parser.add_argument(
            "--end-marker",
            metavar="<end-marker>",
            help=_("End anchor for paging"),
        )
        parser.add_argument(
            "--limit",
            metavar="<num-objects>",
            type=int,
            help=_("Limit the number of objects returned"),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            '--all',
            action='store_true',
            default=False,
            help=_('List all objects in container (default is 10000)'),
        )
        return parser

    def take_action(self, parsed_args):

        if parsed_args.long:
            columns = (
                'Name',
                'Bytes',
                'Hash',
                'Content Type',
                'Last Modified',
            )
        else:
            columns = ('Name',)

        kwargs = {}
        if parsed_args.prefix:
            kwargs['prefix'] = parsed_args.prefix
        if parsed_args.delimiter:
            kwargs['delimiter'] = parsed_args.delimiter
        if parsed_args.marker:
            kwargs['marker'] = parsed_args.marker
        if parsed_args.end_marker:
            kwargs['end_marker'] = parsed_args.end_marker
        if parsed_args.limit:
            kwargs['limit'] = parsed_args.limit
        if parsed_args.all:
            kwargs['full_listing'] = True

        data = self.app.client_manager.object_store.object_list(
            container=parsed_args.container,
            **kwargs
        )

        return (columns,
                (utils.get_dict_properties(
                    s, columns,
                    formatters={},
                ) for s in data))


class SaveObject(command.Command):
    _description = _("Save object locally")

    def get_parser(self, prog_name):
        parser = super(SaveObject, self).get_parser(prog_name)
        parser.add_argument(
            "--file",
            metavar="<filename>",
            help=_("Destination filename (defaults to object name); using '-'"
                   " as the filename will print the file to stdout"),
        )
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Download <object> from <container>'),
        )
        parser.add_argument(
            "object",
            metavar="<object>",
            help=_("Object to save"),
        )
        return parser

    def take_action(self, parsed_args):
        self.app.client_manager.object_store.object_save(
            container=parsed_args.container,
            object=parsed_args.object,
            file=parsed_args.file,
        )


class SetObject(command.Command):
    _description = _("Set object properties")

    def get_parser(self, prog_name):
        parser = super(SetObject, self).get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Modify <object> from <container>'),
        )
        parser.add_argument(
            'object',
            metavar='<object>',
            help=_('Object to modify'),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            required=True,
            action=parseractions.KeyValueAction,
            help=_("Set a property on this object "
                   "(repeat option to set multiple properties)")
        )
        return parser

    def take_action(self, parsed_args):
        self.app.client_manager.object_store.object_set(
            parsed_args.container,
            parsed_args.object,
            properties=parsed_args.property,
        )


class ShowObject(command.ShowOne):
    _description = _("Display object details")

    def get_parser(self, prog_name):
        parser = super(ShowObject, self).get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Display <object> from <container>'),
        )
        parser.add_argument(
            'object',
            metavar='<object>',
            help=_('Object to display'),
        )
        return parser

    def take_action(self, parsed_args):

        data = self.app.client_manager.object_store.object_show(
            container=parsed_args.container,
            object=parsed_args.object,
        )
        if 'properties' in data:
            data['properties'] = format_columns.DictColumn(data['properties'])

        return zip(*sorted(data.items()))


class UnsetObject(command.Command):
    _description = _("Unset object properties")

    def get_parser(self, prog_name):
        parser = super(UnsetObject, self).get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Modify <object> from <container>'),
        )
        parser.add_argument(
            'object',
            metavar='<object>',
            help=_('Object to modify'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            required=True,
            action='append',
            default=[],
            help=_('Property to remove from object '
                   '(repeat option to remove multiple properties)'),
        )
        return parser

    def take_action(self, parsed_args):
        self.app.client_manager.object_store.object_unset(
            parsed_args.container,
            parsed_args.object,
            properties=parsed_args.property,
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import random
import uuid

from openstackclient.tests.functional.network.v2 import common


class SubnetPoolTests(common.NetworkTagTests):
    """Functional tests for subnet pool"""

    base_command = 'subnet pool'

    def setUp(self):
        super(SubnetPoolTests, self).setUp()
        # Nothing in this class works with Nova Network
        if not self.haz_network:
            self.skipTest("No Network service present")

    def test_subnet_pool_create_delete(self):
        """Test create, delete"""
        name1 = uuid.uuid4().hex
        cmd_output, pool_prefix = self._subnet_pool_create("", name1)

        self.assertEqual(
            name1,
            cmd_output["name"]
        )
        self.assertEqual(
            [pool_prefix],
            cmd_output["prefixes"]
        )

        name2 = uuid.uuid4().hex
        cmd_output, pool_prefix = self._subnet_pool_create("", name2)

        self.assertEqual(
            name2,
            cmd_output["name"]
        )
        self.assertEqual(
            [pool_prefix],
            cmd_output["prefixes"]
        )

        del_output = self.openstack(
            'subnet pool delete ' + name1 + ' ' + name2,
        )
        self.assertOutput('', del_output)

    def test_subnet_pool_list(self):
        """Test create, list filter"""
        cmd_output = json.loads(self.openstack('token issue -f json'))
        auth_project_id = cmd_output['project_id']

        cmd_output = json.loads(self.openstack('project list -f json'))
        admin_project_id = None
        demo_project_id = None
        for p in cmd_output:
            if p['Name'] == 'admin':
                admin_project_id = p['ID']
            if p['Name'] == 'demo':
                demo_project_id = p['ID']

        # Verify assumptions:
        # * admin and demo projects are present
        # * demo and admin are distinct projects
        # * tests run as admin
        self.assertIsNotNone(admin_project_id)
        self.assertIsNotNone(demo_project_id)
        self.assertNotEqual(admin_project_id, demo_project_id)
        self.assertEqual(admin_project_id, auth_project_id)

        name1 = uuid.uuid4().hex
        name2 = uuid.uuid4().hex

        cmd_output, pool_prefix = self._subnet_pool_create(
            '--project ' + demo_project_id +
            ' --no-share ',
            name1,
        )
        self.addCleanup(self.openstack, 'subnet pool delete ' + name1)
        self.assertEqual(
            name1,
            cmd_output["name"],
        )
        self.assertEqual(
            False,
            cmd_output["shared"],
        )
        self.assertEqual(
            demo_project_id,
            cmd_output["project_id"],
        )
        self.assertEqual(
            [pool_prefix],
            cmd_output["prefixes"],
        )

        cmd_output, pool_prefix = self._subnet_pool_create(
            ' --share ',
            name2,
        )
        self.addCleanup(self.openstack, 'subnet pool delete ' + name2)
        self.assertEqual(
            name2,
            cmd_output["name"],
        )
        self.assertEqual(
            True,
            cmd_output["shared"],
        )
        self.assertEqual(
            admin_project_id,
            cmd_output["project_id"],
        )
        self.assertEqual(
            [pool_prefix],
            cmd_output["prefixes"],
        )

        # Test list --project
        cmd_output = json.loads(self.openstack(
            'subnet pool list -f json ' +
            '--project ' + demo_project_id
        ))
        names = [x["Name"] for x in cmd_output]
        self.assertIn(name1, names)
        self.assertNotIn(name2, names)

        # Test list --share
        cmd_output = json.loads(self.openstack(
            'subnet pool list -f json ' +
            '--share'
        ))
        names = [x["Name"] for x in cmd_output]
        self.assertNotIn(name1, names)
        self.assertIn(name2, names)

        # Test list --name
        cmd_output = json.loads(self.openstack(
            'subnet pool list -f json ' +
            '--name ' + name1
        ))
        names = [x["Name"] for x in cmd_output]
        self.assertIn(name1, names)
        self.assertNotIn(name2, names)

        # Test list --long
        cmd_output = json.loads(self.openstack(
            'subnet pool list -f json ' +
            '--long '
        ))
        names = [x["Name"] for x in cmd_output]
        self.assertIn(name1, names)
        self.assertIn(name2, names)

    def test_subnet_pool_set_show(self):
        """Test create, delete, set, show, unset"""

        name = uuid.uuid4().hex
        new_name = name + "_"
        cmd_output, pool_prefix = self._subnet_pool_create(
            '--default-prefix-length 16 ' +
            '--min-prefix-length 16 ' +
            '--max-prefix-length 32 ' +
            '--description aaaa ' +
            '--default-quota 10 ',
            name,
        )

        self.addCleanup(
            self.openstack,
            'subnet pool delete ' + cmd_output['id'],
        )
        self.assertEqual(
            name,
            cmd_output["name"],
        )
        self.assertEqual(
            'aaaa',
            cmd_output["description"],
        )
        self.assertEqual(
            [pool_prefix],
            cmd_output["prefixes"],
        )
        self.assertEqual(
            16,
            cmd_output["default_prefixlen"],
        )
        self.assertEqual(
            16,
            cmd_output["min_prefixlen"],
        )
        self.assertEqual(
            32,
            cmd_output["max_prefixlen"],
        )
        self.assertEqual(
            10,
            cmd_output["default_quota"],
        )

        # Test set
        cmd_output = self.openstack(
            'subnet pool set ' +
            '--name ' + new_name +
            ' --description bbbb ' +
            ' --pool-prefix 10.110.0.0/16 ' +
            '--default-prefix-length 8 ' +
            '--min-prefix-length 8 ' +
            '--max-prefix-length 16 ' +
            '--default-quota 20 ' +
            name,
        )
        self.assertOutput('', cmd_output)

        cmd_output = json.loads(self.openstack(
            'subnet pool show -f json ' +
            new_name
        ))
        self.assertEqual(
            new_name,
            cmd_output["name"],
        )
        self.assertEqual(
            'bbbb',
            cmd_output["description"],
        )
        self.assertEqual(
            sorted(["10.110.0.0/16", pool_prefix]),
            sorted(cmd_output["prefixes"]),
        )
        self.assertEqual(
            8,
            cmd_output["default_prefixlen"],
        )
        self.assertEqual(
            8,
            cmd_output["min_prefixlen"],
        )
        self.assertEqual(
            16,
            cmd_output["max_prefixlen"],
        )
        self.assertEqual(
            20,
            cmd_output["default_quota"],
        )

        # Test unset
        # NOTE(dtroyer): The unset command --default-quota option DOES NOT
        #                WORK after a default quota has been set once on a
        #                pool.  The error appears to be in a lower layer,
        #                once that is fixed add a test for subnet pool unset
        #                --default-quota.
        #                The unset command of --pool-prefixes also doesn't work
        #                right now. It would be fixed in a separate patch once
        #                the lower layer is fixed.
        # cmd_output = self.openstack(
        #    '--debug ' +
        #    'subnet pool unset ' +
        #    ' --pool-prefix 10.110.0.0/16 ' +
        #    new_name,
        # )
        # self.assertOutput('', cmd_output)
        # self.assertNone(cmd_output["prefixes"])

    def _subnet_pool_create(self, cmd, name, is_type_ipv4=True):
        """Make a random subnet pool

        :param string cmd:
            The options for a subnet pool create command, not including
            --pool-prefix and <name>
        :param string name:
            The name of the subnet pool
        :param bool is_type_ipv4:
            Creates an IPv4 pool if True, creates an IPv6 pool otherwise

        Try random subnet ranges because we can not determine ahead of time
        what subnets are already in use, possibly by another test running in
        parallel, try 4 times before failing.
        """
        for i in range(4):
            # Create a random prefix
            if is_type_ipv4:
                pool_prefix = ".".join(map(
                    str,
                    (random.randint(0, 223) for _ in range(2)),
                )) + ".0.0/16"
            else:
                pool_prefix = ":".join(map(
                    str,
                    (hex(random.randint(0, 65535))[2:] for _ in range(6)),
                )) + ":0:0/96"

            try:
                cmd_output = json.loads(self.openstack(
                    'subnet pool create -f json ' +
                    cmd + ' ' +
                    '--pool-prefix ' + pool_prefix + ' ' +
                    name
                ))
            except Exception:
                if (i == 3):
                    # Raise the exception the last time
                    raise
                pass
            else:
                # Break and no longer retry if create is successful
                break

        return cmd_output, pool_prefix

    def _create_resource_for_tag_test(self, name, args):
        cmd_output, _pool_prefix = self._subnet_pool_create(args, name)
        return cmd_output



#   Copyright 2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
from unittest import mock

from openstackclient.api import object_store_v1 as object_store
from openstackclient.object.v1 import object as obj
from openstackclient.tests.unit.object.v1 import fakes as object_fakes


AUTH_TOKEN = "foobar"
AUTH_URL = "http://0.0.0.0"


class TestObject(object_fakes.TestObjectv1):

    def setUp(self):
        super(TestObject, self).setUp()
        self.app.client_manager.object_store = object_store.APIv1(
            session=mock.Mock(),
            service_type="object-store",
        )
        self.api = self.app.client_manager.object_store


@mock.patch(
    'openstackclient.api.object_store_v1.APIv1.object_list'
)
class TestObjectList(TestObject):

    columns = ('Name',)
    datalist = (
        (
            object_fakes.object_name_2,
        ),
    )

    def setUp(self):
        super(TestObjectList, self).setUp()

        # Get the command object to test
        self.cmd = obj.ListObject(self.app, None)

    def test_object_list_objects_no_options(self, o_mock):
        o_mock.return_value = [
            copy.deepcopy(object_fakes.OBJECT),
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            object_fakes.container_name,
        ]
        verifylist = [
            ('container', object_fakes.container_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        o_mock.assert_called_with(
            container=object_fakes.container_name,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (object_fakes.object_name_1, ),
            (object_fakes.object_name_2, ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_object_list_objects_prefix(self, o_mock):
        o_mock.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--prefix', 'floppy',
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('prefix', 'floppy'),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'prefix': 'floppy',
        }
        o_mock.assert_called_with(
            container=object_fakes.container_name_2,
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_delimiter(self, o_mock):
        o_mock.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--delimiter', '=',
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('delimiter', '='),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'delimiter': '=',
        }
        o_mock.assert_called_with(
            container=object_fakes.container_name_2,
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_marker(self, o_mock):
        o_mock.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--marker', object_fakes.object_name_2,
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('marker', object_fakes.object_name_2),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'marker': object_fakes.object_name_2,
        }
        o_mock.assert_called_with(
            container=object_fakes.container_name_2,
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_end_marker(self, o_mock):
        o_mock.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--end-marker', object_fakes.object_name_2,
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('end_marker', object_fakes.object_name_2),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'end_marker': object_fakes.object_name_2,
        }
        o_mock.assert_called_with(
            container=object_fakes.container_name_2,
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_limit(self, o_mock):
        o_mock.return_value = [
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--limit', '2',
            object_fakes.container_name_2,
        ]
        verifylist = [
            ('limit', 2),
            ('container', object_fakes.container_name_2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'limit': 2,
        }
        o_mock.assert_called_with(
            container=object_fakes.container_name_2,
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_object_list_objects_long(self, o_mock):
        o_mock.return_value = [
            copy.deepcopy(object_fakes.OBJECT),
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--long',
            object_fakes.container_name,
        ]
        verifylist = [
            ('long', True),
            ('container', object_fakes.container_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
        }
        o_mock.assert_called_with(
            container=object_fakes.container_name,
            **kwargs
        )

        collist = ('Name', 'Bytes', 'Hash', 'Content Type', 'Last Modified')
        self.assertEqual(collist, columns)
        datalist = (
            (
                object_fakes.object_name_1,
                object_fakes.object_bytes_1,
                object_fakes.object_hash_1,
                object_fakes.object_content_type_1,
                object_fakes.object_modified_1,
            ),
            (
                object_fakes.object_name_2,
                object_fakes.object_bytes_2,
                object_fakes.object_hash_2,
                object_fakes.object_content_type_2,
                object_fakes.object_modified_2,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_object_list_objects_all(self, o_mock):
        o_mock.return_value = [
            copy.deepcopy(object_fakes.OBJECT),
            copy.deepcopy(object_fakes.OBJECT_2),
        ]

        arglist = [
            '--all',
            object_fakes.container_name,
        ]
        verifylist = [
            ('all', True),
            ('container', object_fakes.container_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'full_listing': True,
        }
        o_mock.assert_called_with(
            container=object_fakes.container_name,
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (object_fakes.object_name_1, ),
            (object_fakes.object_name_2, ),
        )
        self.assertEqual(datalist, tuple(data))


@mock.patch(
    'openstackclient.api.object_store_v1.APIv1.object_show'
)
class TestObjectShow(TestObject):

    def setUp(self):
        super(TestObjectShow, self).setUp()

        # Get the command object to test
        self.cmd = obj.ShowObject(self.app, None)

    def test_object_show(self, c_mock):
        c_mock.return_value = copy.deepcopy(object_fakes.OBJECT)

        arglist = [
            object_fakes.container_name,
            object_fakes.object_name_1,
        ]
        verifylist = [
            ('container', object_fakes.container_name),
            ('object', object_fakes.object_name_1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
        }
        # lib.container.show_container(api, url, container)
        c_mock.assert_called_with(
            container=object_fakes.container_name,
            object=object_fakes.object_name_1,
            **kwargs
        )

        collist = ('bytes', 'content_type', 'hash', 'last_modified', 'name')
        self.assertEqual(collist, columns)
        datalist = (
            object_fakes.object_bytes_1,
            object_fakes.object_content_type_1,
            object_fakes.object_hash_1,
            object_fakes.object_modified_1,
            object_fakes.object_name_1,
        )
        self.assertEqual(datalist, data)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Role action implementations"""

import logging

from keystoneauth1 import exceptions as ks_exc
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


def _add_identity_and_resource_options_to_parser(parser):
    system_or_domain_or_project = parser.add_mutually_exclusive_group()
    system_or_domain_or_project.add_argument(
        '--system',
        metavar='<system>',
        help=_('Include <system> (all)'),
    )
    system_or_domain_or_project.add_argument(
        '--domain',
        metavar='<domain>',
        help=_('Include <domain> (name or ID)'),
    )
    system_or_domain_or_project.add_argument(
        '--project',
        metavar='<project>',
        help=_('Include <project> (name or ID)'),
    )
    user_or_group = parser.add_mutually_exclusive_group()
    user_or_group.add_argument(
        '--user',
        metavar='<user>',
        help=_('Include <user> (name or ID)'),
    )
    user_or_group.add_argument(
        '--group',
        metavar='<group>',
        help=_('Include <group> (name or ID)'),
    )
    common.add_group_domain_option_to_parser(parser)
    common.add_project_domain_option_to_parser(parser)
    common.add_user_domain_option_to_parser(parser)
    common.add_inherited_option_to_parser(parser)


def _process_identity_and_resource_options(parsed_args,
                                           identity_client_manager,
                                           validate_actor_existence=True):

    def _find_user():
        try:
            return common.find_user(
                identity_client_manager,
                parsed_args.user,
                parsed_args.user_domain
            ).id
        except exceptions.CommandError:
            if not validate_actor_existence:
                return parsed_args.user
            raise

    def _find_group():
        try:
            return common.find_group(
                identity_client_manager,
                parsed_args.group,
                parsed_args.group_domain
            ).id
        except exceptions.CommandError:
            if not validate_actor_existence:
                return parsed_args.group
            raise

    kwargs = {}
    if parsed_args.user and parsed_args.system:
        kwargs['user'] = _find_user()
        kwargs['system'] = parsed_args.system
    elif parsed_args.user and parsed_args.domain:
        kwargs['user'] = _find_user()
        kwargs['domain'] = common.find_domain(
            identity_client_manager,
            parsed_args.domain,
        ).id
    elif parsed_args.user and parsed_args.project:
        kwargs['user'] = _find_user()
        kwargs['project'] = common.find_project(
            identity_client_manager,
            parsed_args.project,
            parsed_args.project_domain,
        ).id
    elif parsed_args.group and parsed_args.system:
        kwargs['group'] = _find_group()
        kwargs['system'] = parsed_args.system
    elif parsed_args.group and parsed_args.domain:
        kwargs['group'] = _find_group()
        kwargs['domain'] = common.find_domain(
            identity_client_manager,
            parsed_args.domain,
        ).id
    elif parsed_args.group and parsed_args.project:
        kwargs['group'] = _find_group()
        kwargs['project'] = common.find_project(
            identity_client_manager,
            parsed_args.project,
            parsed_args.project_domain,
        ).id
    kwargs['os_inherit_extension_inherited'] = parsed_args.inherited
    return kwargs


class AddRole(command.Command):
    _description = _("Adds a role assignment to a user or group on the "
                     "system, a domain, or a project")

    def get_parser(self, prog_name):
        parser = super(AddRole, self).get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to add to <user> (name or ID)'),
        )
        _add_identity_and_resource_options_to_parser(parser)
        common.add_role_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        if (not parsed_args.user and not parsed_args.domain and
                not parsed_args.group and not parsed_args.project):
            msg = _("Role not added, incorrect set of arguments "
                    "provided. See openstack --help for more details")
            raise exceptions.CommandError(msg)

        domain_id = None
        if parsed_args.role_domain:
            domain_id = common.find_domain(identity_client,
                                           parsed_args.role_domain).id
        role = utils.find_resource(
            identity_client.roles,
            parsed_args.role,
            domain_id=domain_id
        )

        kwargs = _process_identity_and_resource_options(
            parsed_args, self.app.client_manager.identity)

        identity_client.roles.grant(role.id, **kwargs)


class CreateRole(command.ShowOne):
    _description = _("Create new role")

    def get_parser(self, prog_name):
        parser = super(CreateRole, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<role-name>',
            help=_('New role name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Add description about the role'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain the role belongs to (name or ID)'),
        )
        parser.add_argument(
            '--or-show',
            action='store_true',
            help=_('Return existing role'),
        )
        common.add_resource_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        domain_id = None
        if parsed_args.domain:
            domain_id = common.find_domain(identity_client,
                                           parsed_args.domain).id

        options = common.get_immutable_options(parsed_args)

        try:
            role = identity_client.roles.create(
                name=parsed_args.name, domain=domain_id,
                description=parsed_args.description, options=options)

        except ks_exc.Conflict:
            if parsed_args.or_show:
                role = utils.find_resource(identity_client.roles,
                                           parsed_args.name,
                                           domain_id=domain_id)
                LOG.info(_('Returning existing role %s'), role.name)
            else:
                raise

        role._info.pop('links')
        return zip(*sorted(role._info.items()))


class DeleteRole(command.Command):
    _description = _("Delete role(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteRole, self).get_parser(prog_name)
        parser.add_argument(
            'roles',
            metavar='<role>',
            nargs="+",
            help=_('Role(s) to delete (name or ID)'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain the role belongs to (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        domain_id = None
        if parsed_args.domain:
            domain_id = common.find_domain(identity_client,
                                           parsed_args.domain).id
        errors = 0
        for role in parsed_args.roles:
            try:
                role_obj = utils.find_resource(
                    identity_client.roles,
                    role,
                    domain_id=domain_id
                )
                identity_client.roles.delete(role_obj.id)
            except Exception as e:
                errors += 1
                LOG.error(_("Failed to delete role with "
                          "name or ID '%(role)s': %(e)s"),
                          {'role': role, 'e': e})

        if errors > 0:
            total = len(parsed_args.roles)
            msg = (_("%(errors)s of %(total)s roles failed "
                   "to delete.") % {'errors': errors, 'total': total})
            raise exceptions.CommandError(msg)


class ListRole(command.Lister):
    _description = _("List roles")

    def get_parser(self, prog_name):
        parser = super(ListRole, self).get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Include <domain> (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        if parsed_args.domain:
            domain = common.find_domain(
                identity_client,
                parsed_args.domain,
            )
            columns = ('ID', 'Name', 'Domain')
            data = identity_client.roles.list(domain_id=domain.id)
            for role in data:
                role.domain = domain.name
        else:
            columns = ('ID', 'Name')
            data = identity_client.roles.list()

        return (columns,
                (utils.get_item_properties(
                    s, columns,
                    formatters={},
                ) for s in data))


class RemoveRole(command.Command):
    _description = _("Removes a role assignment from system/domain/project : "
                     "user/group")

    def get_parser(self, prog_name):
        parser = super(RemoveRole, self).get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to remove (name or ID)'),
        )
        _add_identity_and_resource_options_to_parser(parser)
        common.add_role_domain_option_to_parser(parser)

        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        if (not parsed_args.user and not parsed_args.domain and
                not parsed_args.group and not parsed_args.project):
            msg = _("Incorrect set of arguments provided. "
                    "See openstack --help for more details")
            raise exceptions.CommandError(msg)

        domain_id = None
        if parsed_args.role_domain:
            domain_id = common.find_domain(identity_client,
                                           parsed_args.role_domain).id
        role = utils.find_resource(
            identity_client.roles,
            parsed_args.role,
            domain_id=domain_id
        )

        kwargs = _process_identity_and_resource_options(
            parsed_args, self.app.client_manager.identity,
            validate_actor_existence=False
        )
        identity_client.roles.revoke(role.id, **kwargs)


class SetRole(command.Command):
    _description = _("Set role properties")

    def get_parser(self, prog_name):
        parser = super(SetRole, self).get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to modify (name or ID)'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Add description about the role'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain the role belongs to (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Set role name'),
        )
        common.add_resource_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        domain_id = None
        if parsed_args.domain:
            domain_id = common.find_domain(identity_client,
                                           parsed_args.domain).id

        options = common.get_immutable_options(parsed_args)
        role = utils.find_resource(identity_client.roles,
                                   parsed_args.role,
                                   domain_id=domain_id)

        identity_client.roles.update(role.id, name=parsed_args.name,
                                     description=parsed_args.description,
                                     options=options)


class ShowRole(command.ShowOne):
    _description = _("Display role details")

    def get_parser(self, prog_name):
        parser = super(ShowRole, self).get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to display (name or ID)'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain the role belongs to (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        domain_id = None
        if parsed_args.domain:
            domain_id = common.find_domain(identity_client,
                                           parsed_args.domain).id

        role = utils.find_resource(identity_client.roles,
                                   parsed_args.role,
                                   domain_id=domain_id)

        role._info.pop('links')
        return zip(*sorted(role._info.items()))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Domain action implementations"""

import logging

from keystoneauth1 import exceptions as ks_exc
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


class CreateDomain(command.ShowOne):
    _description = _("Create new domain")

    def get_parser(self, prog_name):
        parser = super(CreateDomain, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<domain-name>',
            help=_('New domain name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New domain description'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            help=_('Enable domain (default)'),
        )
        enable_group.add_argument(
            '--disable',
            action='store_true',
            help=_('Disable domain'),
        )
        parser.add_argument(
            '--or-show',
            action='store_true',
            help=_('Return existing domain'),
        )
        common.add_resource_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        enabled = True
        if parsed_args.disable:
            enabled = False

        options = common.get_immutable_options(parsed_args)

        try:
            domain = identity_client.domains.create(
                name=parsed_args.name,
                description=parsed_args.description,
                options=options,
                enabled=enabled,
            )
        except ks_exc.Conflict:
            if parsed_args.or_show:
                domain = utils.find_resource(identity_client.domains,
                                             parsed_args.name)
                LOG.info(_('Returning existing domain %s'), domain.name)
            else:
                raise

        domain._info.pop('links')
        return zip(*sorted(domain._info.items()))


class DeleteDomain(command.Command):
    _description = _("Delete domain(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteDomain, self).get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            nargs='+',
            help=_('Domain(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        result = 0
        for i in parsed_args.domain:
            try:
                domain = utils.find_resource(identity_client.domains, i)
                identity_client.domains.delete(domain.id)
            except Exception as e:
                result += 1
                LOG.error(_("Failed to delete domain with name or "
                          "ID '%(domain)s': %(e)s"), {'domain': i, 'e': e})

        if result > 0:
            total = len(parsed_args.domain)
            msg = (_("%(result)s of %(total)s domains failed "
                   "to delete.") % {'result': result, 'total': total})
            raise exceptions.CommandError(msg)


class ListDomain(command.Lister):
    _description = _("List domains")

    def get_parser(self, prog_name):
        parser = super(ListDomain, self).get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('The domain name'),
        )
        parser.add_argument(
            '--enabled',
            dest='enabled',
            action='store_true',
            help=_('The domains that are enabled will be returned'),
        )
        return parser

    def take_action(self, parsed_args):
        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.enabled:
            kwargs['enabled'] = True

        columns = ('ID', 'Name', 'Enabled', 'Description')
        data = self.app.client_manager.identity.domains.list(**kwargs)
        return (columns,
                (utils.get_item_properties(
                    s, columns,
                    formatters={},
                ) for s in data))


class SetDomain(command.Command):
    _description = _("Set domain properties")

    def get_parser(self, prog_name):
        parser = super(SetDomain, self).get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help=_('Domain to modify (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New domain name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New domain description'),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            help=_('Enable domain'),
        )
        enable_group.add_argument(
            '--disable',
            action='store_true',
            help=_('Disable domain'),
        )
        common.add_resource_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        domain = utils.find_resource(identity_client.domains,
                                     parsed_args.domain)
        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.description:
            kwargs['description'] = parsed_args.description

        if parsed_args.enable:
            kwargs['enabled'] = True
        if parsed_args.disable:
            kwargs['enabled'] = False

        options = common.get_immutable_options(parsed_args)
        if options:
            kwargs['options'] = options

        identity_client.domains.update(domain.id, **kwargs)


class ShowDomain(command.ShowOne):
    _description = _("Display domain details")

    def get_parser(self, prog_name):
        parser = super(ShowDomain, self).get_parser(prog_name)
        parser.add_argument(
            'domain',
            metavar='<domain>',
            help=_('Domain to display (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        domain_str = common._get_token_resource(identity_client, 'domain',
                                                parsed_args.domain)

        domain = utils.find_resource(identity_client.domains,
                                     domain_str)

        domain._info.pop('links')
        return zip(*sorted(domain._info.items()))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Manage access to the clients, including authenticating when needed."""

import importlib
import logging
import sys

from osc_lib import clientmanager
from osc_lib import shell
import stevedore


LOG = logging.getLogger(__name__)

PLUGIN_MODULES = []

USER_AGENT = 'python-openstackclient'


class ClientManager(clientmanager.ClientManager):
    """Manages access to API clients, including authentication

    Wrap osc_lib's ClientManager to maintain compatibility for the existing
    plugin V2 interface.  Some currently private attributes become public
    in osc-lib so we need to maintain a transition period.
    """

    # A simple incrementing version for the plugin to know what is available
    PLUGIN_INTERFACE_VERSION = "2"

    # Let the commands set this
    _auth_required = False

    def __init__(
        self,
        cli_options=None,
        api_version=None,
        pw_func=None,
    ):
        super(ClientManager, self).__init__(
            cli_options=cli_options,
            api_version=api_version,
            pw_func=pw_func,
        )

        # TODO(dtroyer): For compatibility; mark this for removal when plugin
        #                interface v2 is removed
        self._region_name = self.region_name
        self._interface = self.interface
        self._cacert = self.cacert
        self._insecure = not self.verify
        # store original auth_type
        self._original_auth_type = cli_options.auth_type

    def setup_auth(self):
        """Set up authentication"""

        if self._auth_setup_completed:
            return

        # NOTE(dtroyer): Validate the auth args; this is protected with 'if'
        #                because openstack_config is an optional argument to
        #                CloudConfig.__init__() and we'll die if it was not
        #                passed.
        if (
                self._auth_required and
                self._cli_options._openstack_config is not None
        ):
            self._cli_options._openstack_config._pw_callback = \
                shell.prompt_for_password
            try:
                # We might already get auth from SDK caching
                if not self._cli_options._auth:
                    self._cli_options._auth = \
                        self._cli_options._openstack_config.load_auth_plugin(
                            self._cli_options.config,
                        )
            except TypeError as e:
                self._fallback_load_auth_plugin(e)

        return super(ClientManager, self).setup_auth()

    def _fallback_load_auth_plugin(self, e):
        # NOTES(RuiChen): Hack to avoid auth plugins choking on data they don't
        #                 expect, delete fake token and endpoint, then try to
        #                 load auth plugin again with user specified options.
        #                 We know it looks ugly, but it's necessary.
        if self._cli_options.config['auth']['token'] == 'x':
            # restore original auth_type
            self._cli_options.config['auth_type'] = \
                self._original_auth_type
            del self._cli_options.config['auth']['token']
            del self._cli_options.config['auth']['endpoint']
            self._cli_options._auth = \
                self._cli_options._openstack_config.load_auth_plugin(
                    self._cli_options.config,
                )
        else:
            raise e

    def is_network_endpoint_enabled(self):
        """Check if the network endpoint is enabled"""

        # NOTE(dtroyer): is_service_available() can also return None if
        #                there is no Service Catalog, callers here are
        #                not expecting that so fold None into True to
        #                use Network API by default
        return self.is_service_available('network') is not False

    def is_compute_endpoint_enabled(self):
        """Check if Compute endpoint is enabled"""

        return self.is_service_available('compute') is not False

    def is_volume_endpoint_enabled(self, volume_client):
        """Check if volume endpoint is enabled"""
        # NOTE(jcross): Cinder did some interesting things with their service
        #               name so we need to figure out which version to look
        #               for when calling is_service_available()
        volume_version = volume_client.api_version.ver_major
        if self.is_service_available(
                "volumev%s" % volume_version) is not False:
            return True
        elif self.is_service_available('volume') is not False:
            return True
        else:
            return False


# Plugin Support

def get_plugin_modules(group):
    """Find plugin entry points"""
    mod_list = []
    mgr = stevedore.ExtensionManager(group)
    for ep in mgr:
        LOG.debug('Found plugin %s', ep.name)

        # Different versions of stevedore use different
        # implementations of EntryPoint from other libraries, which
        # are not API-compatible.
        try:
            module_name = ep.entry_point.module_name
        except AttributeError:
            try:
                module_name = ep.entry_point.module
            except AttributeError:
                module_name = ep.entry_point.value

        try:
            module = importlib.import_module(module_name)
        except Exception as err:
            sys.stderr.write(
                "WARNING: Failed to import plugin %s: %s.\n" % (ep.name, err))
            continue

        mod_list.append(module)
        init_func = getattr(module, 'Initialize', None)
        if init_func:
            init_func('x')

        # Add the plugin to the ClientManager
        setattr(
            clientmanager.ClientManager,
            module.API_NAME,
            clientmanager.ClientCache(
                getattr(sys.modules[module_name], 'make_client', None)
            ),
        )
    return mod_list


def build_plugin_option_parser(parser):
    """Add plugin options to the parser"""

    # Loop through extensions to get parser additions
    for mod in PLUGIN_MODULES:
        parser = mod.build_option_parser(parser)
    return parser


# Get list of base plugin modules
PLUGIN_MODULES = get_plugin_modules(
    'openstack.cli.base',
)
# Append list of external plugin modules
PLUGIN_MODULES.extend(get_plugin_modules(
    'openstack.cli.extension',
))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock
from unittest.mock import call

from osc_lib import exceptions

from openstackclient.identity.v3 import credential
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit import utils


class TestCredential(identity_fakes.TestIdentityv3):

    def setUp(self):
        super(TestCredential, self).setUp()

        # Get a shortcut to the CredentialManager Mock
        self.credentials_mock = self.app.client_manager.identity.credentials
        self.credentials_mock.reset_mock()

        # Get a shortcut to the UserManager Mock
        self.users_mock = self.app.client_manager.identity.users
        self.users_mock.reset_mock()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.app.client_manager.identity.projects
        self.projects_mock.reset_mock()


class TestCredentialCreate(TestCredential):

    user = identity_fakes.FakeUser.create_one_user()
    project = identity_fakes.FakeProject.create_one_project()
    columns = (
        'blob',
        'id',
        'project_id',
        'type',
        'user_id',
    )

    def setUp(self):
        super(TestCredentialCreate, self).setUp()

        self.credential = identity_fakes.FakeCredential.create_one_credential(
            attrs={'user_id': self.user.id, 'project_id': self.project.id})
        self.credentials_mock.create.return_value = self.credential
        self.users_mock.get.return_value = self.user
        self.projects_mock.get.return_value = self.project
        self.data = (
            self.credential.blob,
            self.credential.id,
            self.credential.project_id,
            self.credential.type,
            self.credential.user_id,
        )

        self.cmd = credential.CreateCredential(self.app, None)

    def test_credential_create_no_options(self):
        arglist = [
            self.credential.user_id,
            self.credential.blob,
        ]
        verifylist = [
            ('user', self.credential.user_id),
            ('data', self.credential.blob),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'user': self.credential.user_id,
            'type': self.credential.type,
            'blob': self.credential.blob,
            'project': None,
        }
        self.credentials_mock.create.assert_called_once_with(
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_credential_create_with_options(self):
        arglist = [
            self.credential.user_id,
            self.credential.blob,
            '--type', self.credential.type,
            '--project', self.credential.project_id,
        ]
        verifylist = [
            ('user', self.credential.user_id),
            ('data', self.credential.blob),
            ('type', self.credential.type),
            ('project', self.credential.project_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'user': self.credential.user_id,
            'type': self.credential.type,
            'blob': self.credential.blob,
            'project': self.credential.project_id,
        }
        self.credentials_mock.create.assert_called_once_with(
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestCredentialDelete(TestCredential):

    credentials = identity_fakes.FakeCredential.create_credentials(count=2)

    def setUp(self):
        super(TestCredentialDelete, self).setUp()

        self.credentials_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = credential.DeleteCredential(self.app, None)

    def test_credential_delete(self):
        arglist = [
            self.credentials[0].id,
        ]
        verifylist = [
            ('credential', [self.credentials[0].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.credentials_mock.delete.assert_called_with(
            self.credentials[0].id,
        )
        self.assertIsNone(result)

    def test_credential_multi_delete(self):
        arglist = []
        for c in self.credentials:
            arglist.append(c.id)
        verifylist = [
            ('credential', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for c in self.credentials:
            calls.append(call(c.id))
        self.credentials_mock.delete.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_credential_multi_delete_with_exception(self):
        arglist = [
            self.credentials[0].id,
            'unexist_credential',
        ]
        verifylist = [
            ('credential', [self.credentials[0].id, 'unexist_credential'])
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        delete_mock_result = [None, exceptions.CommandError]
        self.credentials_mock.delete = (
            mock.Mock(side_effect=delete_mock_result)
        )

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 credential failed to delete.', str(e))

        self.credentials_mock.delete.assert_any_call(self.credentials[0].id)
        self.credentials_mock.delete.assert_any_call('unexist_credential')


class TestCredentialList(TestCredential):

    credential = identity_fakes.FakeCredential.create_one_credential()

    columns = ('ID', 'Type', 'User ID', 'Data', 'Project ID')
    data = ((
        credential.id,
        credential.type,
        credential.user_id,
        credential.blob,
        credential.project_id,
    ), )

    def setUp(self):
        super(TestCredentialList, self).setUp()

        self.user = identity_fakes.FakeUser.create_one_user()
        self.users_mock.get.return_value = self.user

        self.credentials_mock.list.return_value = [self.credential]

        # Get the command object to test
        self.cmd = credential.ListCredential(self.app, None)

    def test_credential_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.credentials_mock.list.assert_called_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

    def test_credential_list_with_options(self):
        arglist = [
            '--user', self.credential.user_id,
            '--type', self.credential.type,
        ]
        verifylist = [
            ('user', self.credential.user_id),
            ('type', self.credential.type),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'user_id': self.user.id,
            'type': self.credential.type,
        }
        self.users_mock.get.assert_called_with(self.credential.user_id)
        self.credentials_mock.list.assert_called_with(
            **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))


class TestCredentialSet(TestCredential):

    credential = identity_fakes.FakeCredential.create_one_credential()

    def setUp(self):
        super(TestCredentialSet, self).setUp()
        self.cmd = credential.SetCredential(self.app, None)

    def test_credential_set_no_options(self):
        arglist = [
            self.credential.id,
        ]

        self.assertRaises(utils.ParserException,
                          self.check_parser, self.cmd, arglist, [])

    def test_credential_set_missing_user(self):
        arglist = [
            '--type', 'ec2',
            '--data', self.credential.blob,
            self.credential.id,
        ]

        self.assertRaises(utils.ParserException,
                          self.check_parser, self.cmd, arglist, [])

    def test_credential_set_missing_type(self):
        arglist = [
            '--user', self.credential.user_id,
            '--data', self.credential.blob,
            self.credential.id,
        ]

        self.assertRaises(utils.ParserException,
                          self.check_parser, self.cmd, arglist, [])

    def test_credential_set_missing_data(self):
        arglist = [
            '--user', self.credential.user_id,
            '--type', 'ec2',
            self.credential.id,
        ]

        self.assertRaises(utils.ParserException,
                          self.check_parser, self.cmd, arglist, [])

    def test_credential_set_valid(self):
        arglist = [
            '--user', self.credential.user_id,
            '--type', 'ec2',
            '--data', self.credential.blob,
            self.credential.id,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

    def test_credential_set_valid_with_project(self):
        arglist = [
            '--user', self.credential.user_id,
            '--type', 'ec2',
            '--data', self.credential.blob,
            '--project', self.credential.project_id,
            self.credential.id,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)


class TestCredentialShow(TestCredential):

    columns = (
        'blob',
        'id',
        'project_id',
        'type',
        'user_id',
    )

    def setUp(self):
        super(TestCredentialShow, self).setUp()

        self.credential = identity_fakes.FakeCredential.create_one_credential()
        self.credentials_mock.get.return_value = self.credential
        self.data = (
            self.credential.blob,
            self.credential.id,
            self.credential.project_id,
            self.credential.type,
            self.credential.user_id,
        )

        self.cmd = credential.ShowCredential(self.app, None)

    def test_credential_show(self):
        arglist = [
            self.credential.id,
        ]
        verifylist = [
            ('credential', self.credential.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.credentials_mock.get.assert_called_once_with(self.credential.id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



# Copyright (c) 2019, Intel Corporation.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import json
import uuid

from openstackclient.tests.functional.network.v2 import common


class NetworkSegmentRangeTests(common.NetworkTests):
    """Functional tests for network segment range"""

    def setUp(self):
        super(NetworkSegmentRangeTests, self).setUp()
        # Nothing in this class works with Nova Network
        if not self.haz_network:
            self.skipTest("No Network service present")
        if not self.is_extension_enabled('network-segment-range'):
            self.skipTest("No network-segment-range extension present")
        self.PROJECT_NAME = uuid.uuid4().hex

    def test_network_segment_range_create_delete(self):
        # Make a project
        project_id = json.loads(self.openstack(
            'project create -f json ' + self.PROJECT_NAME))['id']
        name = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            ' network segment range create -f json ' +
            '--private ' +
            "--project " + self.PROJECT_NAME + " " +
            '--network-type vxlan ' +
            '--minimum 2005 ' +
            '--maximum 2009 ' +
            name
        ))
        self.assertEqual(
            name,
            json_output["name"],
        )
        self.assertEqual(
            project_id,
            json_output["project_id"],
        )

        raw_output = self.openstack(
            'network segment range delete ' + name,
        )
        self.assertOutput('', raw_output)
        raw_output_project = self.openstack(
            'project delete ' + self.PROJECT_NAME)
        self.assertEqual('', raw_output_project)

    def test_network_segment_range_list(self):
        name = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            ' network segment range create -f json ' +
            '--shared ' +
            '--network-type geneve ' +
            '--minimum 2013 ' +
            '--maximum 2017 ' +
            name
        ))
        network_segment_range_id = json_output.get('id')
        network_segment_range_name = json_output.get('name')
        self.addCleanup(
            self.openstack,
            'network segment range delete ' + network_segment_range_id
        )
        self.assertEqual(
            name,
            json_output["name"],
        )

        json_output = json.loads(self.openstack(
            'network segment range list -f json'
        ))
        item_map = {
            item.get('ID'): item.get('Name') for item in json_output
        }
        self.assertIn(network_segment_range_id, item_map.keys())
        self.assertIn(network_segment_range_name, item_map.values())

    def test_network_segment_range_set_show(self):
        project_id = json.loads(self.openstack(
            'project create -f json ' + self.PROJECT_NAME))['id']
        name = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            ' network segment range create -f json ' +
            '--private ' +
            "--project " + self.PROJECT_NAME + " " +
            '--network-type geneve ' +
            '--minimum 2021 ' +
            '--maximum 2025 ' +
            name
        ))
        self.addCleanup(
            self.openstack,
            'network segment range delete ' + name
        )
        self.assertEqual(
            name,
            json_output["name"],
        )
        self.assertEqual(
            project_id,
            json_output["project_id"],
        )

        new_minimum = 2020
        new_maximum = 2029
        cmd_output = self.openstack(
            'network segment range set --minimum {min} --maximum {max} {name}'
            .format(min=new_minimum, max=new_maximum, name=name)
        )
        self.assertOutput('', cmd_output)

        json_output = json.loads(self.openstack(
            'network segment range show -f json ' +
            name
        ))
        self.assertEqual(
            new_minimum,
            json_output["minimum"],
        )
        self.assertEqual(
            new_maximum,
            json_output["maximum"],
        )

        raw_output_project = self.openstack(
            'project delete ' + self.PROJECT_NAME)
        self.assertEqual('', raw_output_project)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from openstackclient.tests.functional.volume.v2 import common


class VolumeServiceTests(common.BaseVolumeTests):
    """Functional tests for volume service."""

    def test_volume_service_list(self):
        cmd_output = json.loads(self.openstack(
            'volume service list -f json'))

        # Get the nonredundant services and hosts
        services = list(set([x['Binary'] for x in cmd_output]))
        hosts = list(set([x['Host'] for x in cmd_output]))

        # Test volume service list --service
        cmd_output = json.loads(self.openstack(
            'volume service list -f json ' +
            '--service ' +
            services[0]
        ))
        for x in cmd_output:
            self.assertEqual(
                services[0],
                x['Binary']
            )

        # Test volume service list --host
        cmd_output = json.loads(self.openstack(
            'volume service list -f json ' +
            '--host ' +
            hosts[0]
        ))
        for x in cmd_output:
            self.assertIn(
                hosts[0],
                x['Host']
            )

    def test_volume_service_set(self):

        # Get a service and host
        cmd_output = json.loads(self.openstack(
            'volume service list -f json'
        ))
        service_1 = cmd_output[0]['Binary']
        host_1 = cmd_output[0]['Host']

        # Test volume service set --enable
        raw_output = self.openstack(
            'volume service set --enable ' +
            host_1 + ' ' +
            service_1
        )
        self.assertOutput('', raw_output)

        cmd_output = json.loads(self.openstack(
            'volume service list -f json --long'
        ))
        self.assertEqual(
            'enabled',
            cmd_output[0]['Status']
        )
        self.assertIsNone(
            cmd_output[0]['Disabled Reason']
        )

        # Test volume service set --disable and --disable-reason
        disable_reason = 'disable_reason'
        raw_output = self.openstack(
            'volume service set --disable ' +
            '--disable-reason ' +
            disable_reason + ' ' +
            host_1 + ' ' +
            service_1
        )
        self.assertOutput('', raw_output)

        cmd_output = json.loads(self.openstack(
            'volume service list -f json --long'
        ))
        self.assertEqual(
            'disabled',
            cmd_output[0]['Status']
        )
        self.assertEqual(
            disable_reason,
            cmd_output[0]['Disabled Reason']
        )



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume v2 snapshot action implementations"""

import copy
import functools
import logging

from cliff import columns as cliff_columns
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)


class VolumeIdColumn(cliff_columns.FormattableColumn):
    """Formattable column for volume ID column.

    Unlike the parent FormattableColumn class, the initializer of the
    class takes volume_cache as the second argument.
    osc_lib.utils.get_item_properties instantiate cliff FormattableColumn
    object with a single parameter "column value", so you need to pass
    a partially initialized class like
    ``functools.partial(VolumeIdColumn, volume_cache)``.
    """

    def __init__(self, value, volume_cache=None):
        super(VolumeIdColumn, self).__init__(value)
        self._volume_cache = volume_cache or {}

    def human_readable(self):
        """Return a volume name if available

        :rtype: either the volume ID or name
        """
        volume_id = self._value
        volume = volume_id
        if volume_id in self._volume_cache.keys():
            volume = self._volume_cache[volume_id].name
        return volume


class CreateVolumeSnapshot(command.ShowOne):
    _description = _("Create new volume snapshot")

    def get_parser(self, prog_name):
        parser = super(CreateVolumeSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            "snapshot_name",
            metavar="<snapshot-name>",
            help=_("Name of the new snapshot"),
        )
        parser.add_argument(
            "--volume",
            metavar="<volume>",
            help=_("Volume to snapshot (name or ID) "
                   "(default is <snapshot-name>)")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Description of the snapshot")
        )
        parser.add_argument(
            "--force",
            action="store_true",
            default=False,
            help=_("Create a snapshot attached to an instance. "
                   "Default is False")
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            help=_("Set a property to this snapshot "
                   "(repeat option to set multiple properties)"),
        )
        parser.add_argument(
            "--remote-source",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            help=_("The attribute(s) of the existing remote volume snapshot "
                   "(admin required) (repeat option to specify multiple "
                   "attributes) e.g.: '--remote-source source-name=test_name "
                   "--remote-source source-id=test_id'"),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume = parsed_args.volume
        if not parsed_args.volume:
            volume = parsed_args.snapshot_name
        volume_id = utils.find_resource(
            volume_client.volumes, volume).id
        if parsed_args.remote_source:
            # Create a new snapshot from an existing remote snapshot source
            if parsed_args.force:
                msg = (_("'--force' option will not work when you create "
                         "new volume snapshot from an existing remote "
                         "volume snapshot"))
                LOG.warning(msg)
            snapshot = volume_client.volume_snapshots.manage(
                volume_id=volume_id,
                ref=parsed_args.remote_source,
                name=parsed_args.snapshot_name,
                description=parsed_args.description,
                metadata=parsed_args.property,
            )
        else:
            # create a new snapshot from scratch
            snapshot = volume_client.volume_snapshots.create(
                volume_id,
                force=parsed_args.force,
                name=parsed_args.snapshot_name,
                description=parsed_args.description,
                metadata=parsed_args.property,
            )
        snapshot._info.update(
            {'properties':
             format_columns.DictColumn(snapshot._info.pop('metadata'))}
        )
        return zip(*sorted(snapshot._info.items()))


class DeleteVolumeSnapshot(command.Command):
    _description = _("Delete volume snapshot(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteVolumeSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            "snapshots",
            metavar="<snapshot>",
            nargs="+",
            help=_("Snapshot(s) to delete (name or ID)")
        )
        parser.add_argument(
            '--force',
            action='store_true',
            help=_("Attempt forced removal of snapshot(s), "
                   "regardless of state (defaults to False)")
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        result = 0

        for i in parsed_args.snapshots:
            try:
                snapshot_id = utils.find_resource(
                    volume_client.volume_snapshots, i).id
                volume_client.volume_snapshots.delete(
                    snapshot_id, parsed_args.force)
            except Exception as e:
                result += 1
                LOG.error(_("Failed to delete snapshot with "
                            "name or ID '%(snapshot)s': %(e)s")
                          % {'snapshot': i, 'e': e})

        if result > 0:
            total = len(parsed_args.snapshots)
            msg = (_("%(result)s of %(total)s snapshots failed "
                   "to delete.") % {'result': result, 'total': total})
            raise exceptions.CommandError(msg)


class ListVolumeSnapshot(command.Lister):
    _description = _("List volume snapshots")

    def get_parser(self, prog_name):
        parser = super(ListVolumeSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Filter results by project (name or ID) (admin only)')
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            '--marker',
            metavar='<volume-snapshot>',
            help=_('The last snapshot ID of the previous page'),
        )
        parser.add_argument(
            '--limit',
            type=int,
            action=parseractions.NonNegativeAction,
            metavar='<num-snapshots>',
            help=_('Maximum number of snapshots to display'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help=_('Filters results by a name.')
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            choices=['available', 'error', 'creating', 'deleting',
                     'error_deleting'],
            help=_("Filters results by a status. "
                   "('available', 'error', 'creating', 'deleting'"
                   " or 'error_deleting')")
        )
        parser.add_argument(
            '--volume',
            metavar='<volume>',
            default=None,
            help=_('Filters results by a volume (name or ID).')
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        identity_client = self.app.client_manager.identity

        if parsed_args.long:
            columns = ['ID', 'Name', 'Description', 'Status',
                       'Size', 'Created At', 'Volume ID', 'Metadata']
            column_headers = copy.deepcopy(columns)
            column_headers[6] = 'Volume'
            column_headers[7] = 'Properties'
        else:
            columns = ['ID', 'Name', 'Description', 'Status', 'Size']
            column_headers = copy.deepcopy(columns)

        # Cache the volume list
        volume_cache = {}
        try:
            for s in volume_client.volumes.list():
                volume_cache[s.id] = s
        except Exception:
            # Just forget it if there's any trouble
            pass
        _VolumeIdColumn = functools.partial(VolumeIdColumn,
                                            volume_cache=volume_cache)

        volume_id = None
        if parsed_args.volume:
            volume_id = utils.find_resource(
                volume_client.volumes, parsed_args.volume).id

        project_id = None
        if parsed_args.project:
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain).id

        # set value of 'all_tenants' when using project option
        all_projects = True if parsed_args.project else \
            parsed_args.all_projects

        search_opts = {
            'all_tenants': all_projects,
            'project_id': project_id,
            'name': parsed_args.name,
            'status': parsed_args.status,
            'volume_id': volume_id,
        }

        data = volume_client.volume_snapshots.list(
            search_opts=search_opts,
            marker=parsed_args.marker,
            limit=parsed_args.limit,
        )
        return (column_headers,
                (utils.get_item_properties(
                    s, columns,
                    formatters={'Metadata': format_columns.DictColumn,
                                'Volume ID': _VolumeIdColumn},
                ) for s in data))


class SetVolumeSnapshot(command.Command):
    _description = _("Set volume snapshot properties")

    def get_parser(self, prog_name):
        parser = super(SetVolumeSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar='<snapshot>',
            help=_('Snapshot to modify (name or ID)')
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New snapshot name')
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New snapshot description')
        )
        parser.add_argument(
            "--no-property",
            dest="no_property",
            action="store_true",
            help=_("Remove all properties from <snapshot> "
                   "(specify both --no-property and --property to "
                   "remove the current properties before setting "
                   "new properties.)"),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            help=_('Property to add/change for this snapshot '
                   '(repeat option to set multiple properties)'),
        )
        parser.add_argument(
            '--state',
            metavar='<state>',
            choices=['available', 'error', 'creating', 'deleting',
                     'error_deleting'],
            help=_('New snapshot state. ("available", "error", "creating", '
                   '"deleting", or "error_deleting") (admin only) '
                   '(This option simply changes the state of the snapshot '
                   'in the database with no regard to actual status, '
                   'exercise caution when using)'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        snapshot = utils.find_resource(volume_client.volume_snapshots,
                                       parsed_args.snapshot)

        result = 0
        if parsed_args.no_property:
            try:
                key_list = snapshot.metadata.keys()
                volume_client.volume_snapshots.delete_metadata(
                    snapshot.id,
                    list(key_list),
                )
            except Exception as e:
                LOG.error(_("Failed to clean snapshot properties: %s"), e)
                result += 1

        if parsed_args.property:
            try:
                volume_client.volume_snapshots.set_metadata(
                    snapshot.id, parsed_args.property)
            except Exception as e:
                LOG.error(_("Failed to set snapshot property: %s"), e)
                result += 1

        if parsed_args.state:
            try:
                volume_client.volume_snapshots.reset_state(
                    snapshot.id, parsed_args.state)
            except Exception as e:
                LOG.error(_("Failed to set snapshot state: %s"), e)
                result += 1

        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.description:
            kwargs['description'] = parsed_args.description
        if kwargs:
            try:
                volume_client.volume_snapshots.update(
                    snapshot.id, **kwargs)
            except Exception as e:
                LOG.error(_("Failed to update snapshot name "
                          "or description: %s"), e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(_("One or more of the "
                                          "set operations failed"))


class ShowVolumeSnapshot(command.ShowOne):
    _description = _("Display volume snapshot details")

    def get_parser(self, prog_name):
        parser = super(ShowVolumeSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_("Snapshot to display (name or ID)")
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        snapshot = utils.find_resource(
            volume_client.volume_snapshots, parsed_args.snapshot)
        snapshot._info.update(
            {'properties':
             format_columns.DictColumn(snapshot._info.pop('metadata'))}
        )
        return zip(*sorted(snapshot._info.items()))


class UnsetVolumeSnapshot(command.Command):
    _description = _("Unset volume snapshot properties")

    def get_parser(self, prog_name):
        parser = super(UnsetVolumeSnapshot, self).get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar='<snapshot>',
            help=_('Snapshot to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            default=[],
            help=_('Property to remove from snapshot '
                   '(repeat option to remove multiple properties)'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        snapshot = utils.find_resource(
            volume_client.volume_snapshots, parsed_args.snapshot)

        if parsed_args.property:
            volume_client.volume_snapshots.delete_metadata(
                snapshot.id,
                parsed_args.property,
            )
